# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Harmonica algorithm for boolean search spaces from 'Hyperparameter Optimization: A Spectral Approach' (https://arxiv.org/abs/1706.00764).

This is a faithful re-implementation based off
https://github.com/callowbird/Harmonica.
"""
# pylint:disable=invalid-name
import functools
import itertools
from typing import Callable, Optional, Sequence, Set
import numpy as np
from sklearn import linear_model
from sklearn import preprocessing

from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier._src.algorithms.designers import random


def _binary_subset_enumeration(dim: int,
                               indices: Sequence[int],
                               default_value: float = 1.0) -> np.ndarray:
  """Outputs all possible binary vectors from {-1, 1}^{dim} where only positions from `indices` are changed."""
  output = default_value * np.ones(
      shape=(2**len(indices), dim), dtype=np.float32)
  for i, binary in enumerate(
      itertools.product([-1.0, 1.0], repeat=len(indices))):
    output[i, indices] = binary
  return output


# TODO: Implement __repr__ via `attr`.
class PolynomialSparseRecovery:
  """Performs LASSO regression over low (d) degree polynomial coefficients."""

  def __init__(self,
               d: int = 3,
               num_top_monomials: int = 5,
               alpha: float = 3.0):
    """Init.

    Args:
      d: Maximum degree of monomials to use during polynomial regression.
      num_top_monomials: Number of top monomial coefficients to use.
      alpha: LASSO regularization coefficient.
    """
    self._d = d
    self._num_top_monomials = num_top_monomials
    self._alpha = alpha
    self.reset()

  def reset(self):
    self._poly_transformer = preprocessing.PolynomialFeatures(
        self._d, interaction_only=True)
    self._top_poly_indices: Optional[np.ndarray] = None
    self._top_poly_coefficients: Optional[np.ndarray] = None

  def regress(self, X: np.ndarray, Y: np.ndarray) -> None:
    """Performs LASSO regression to obtain top monomial coefficients."""

    # Computes monomial values for every vector in X.
    X_poly_features = []
    for x_vector in X:
      X_poly_features.append(
          self._poly_transformer.fit_transform(
              np.array(x_vector).reshape(1, -1))[0].tolist())
    X_poly_features = np.array(X_poly_features)

    # Find optimial coefficients on monomials to the data.
    lasso_solver = linear_model.Lasso(fit_intercept=True, alpha=self._alpha)
    lasso_solver.fit(X_poly_features, Y)

    # Sort and obtain top coefficients.
    lasso_coefficients = lasso_solver.coef_
    poly_indices = np.argsort(-np.abs(lasso_coefficients))
    self._top_poly_indices = poly_indices[:self._num_top_monomials]
    self._top_poly_coefficients = lasso_coefficients[self._top_poly_indices]

  def surrogate(self, x: np.ndarray) -> float:
    """Surrogate/Predicted function after regress()."""
    x_poly_features = self._poly_transformer.fit_transform(x.reshape(1, -1))
    top_x_poly_features = np.take_along_axis(
        x_poly_features[0], self._top_poly_indices, axis=0)
    return np.dot(top_x_poly_features, self._top_poly_coefficients)

  def index_set(self) -> Set[int]:
    """Returns parameter index set J from top coefficients.

    For example, if our monomials corresponding to top cofficients were x0*x1,
    x3, x1*x2*x3, then the output would be Union({0, 1}, {3}, {1, 2, 3}).
    """
    index_set = set()
    poly_feature_one_hots = self._poly_transformer.powers_
    for top_poly_index in self._top_poly_indices:
      active_indices = np.nonzero(
          poly_feature_one_hots[top_poly_index])[0].tolist()
      index_set = index_set | set(active_indices)
    return index_set


def _restricted_surrogate(x: np.ndarray, X_restrictors: np.ndarray,
                          replacement_indices: Sequence[int],
                          psr: PolynomialSparseRecovery):
  """New surrogate with input x's positions replaced from X_restrictor values."""
  objectives = []
  for x_restrictor in X_restrictors:
    x_copy = np.copy(x)
    x_copy[replacement_indices] = x_restrictor[replacement_indices]
    objectives.append(psr.surrogate(x_copy))
  return np.mean(objectives)


# TODO: Implement __repr__ via `attr`.
class HarmonicaQ:
  """Q-stage Harmonica.

  At each stage:
  1. Invoke PSR on the previous data (X,Y).
  2. Obtain t maximizers of the surrogate of the PSR.
  3. Redefine a 'restricted surrogate' using the t maximizers.
  4. Produce a new (X', Y') dataset via random search on this 'restricted
  surrogate'.
  """

  def __init__(self,
               psr: Optional[PolynomialSparseRecovery] = None,
               q: int = 10,
               t: int = 1,
               T: int = 300):
    """Init.

    Args:
      psr: PolynomialSparseRecovery class. If None, will be constructed with
        default arguments.
      q: Number of stages.
      t: Number of maximizers on the surrogate to use.
      T: Number of data samples to collect on the restricted surrogate.
    """
    self._psr = psr
    self._psr = psr or PolynomialSparseRecovery()
    self._q = q
    self._t = t
    self._T = T
    self.reset()

  def reset(self):
    self._restricted_surrogate: Optional[Callable[[np.ndarray], float]] = None
    self._psr.reset()

  def regress(self, X: np.ndarray, Y: np.ndarray) -> None:
    """Performs q-stage Harmonica."""
    num_vars = X.shape[-1]

    X_temp = X
    Y_temp = Y
    for _ in range(self._q):
      # Invoke PSR on data.
      self._psr.reset()
      self._psr.regress(X_temp, Y_temp)
      J = self._psr.index_set()

      # Perform brute force maximization to optain top t optimizers.
      all_X_in_J = _binary_subset_enumeration(num_vars, list(J))
      all_Y_in_J = np.array([self._psr.surrogate(x) for x in all_X_in_J])
      maximizer_idxs = np.argsort(all_Y_in_J)[-self._t:]
      X_maximizers = all_X_in_J[maximizer_idxs]

      # Define restricted surrogate and obtain data from it.
      self._restricted_surrogate = functools.partial(
          _restricted_surrogate,
          X_restrictors=X_maximizers,
          replacement_indices=list(J),
          psr=self._psr)
      X_temp = np.random.choice([-1.0, 1.0], size=(self._T, num_vars))
      Y_temp = np.array([self._restricted_surrogate(x) for x in X_temp])

  def surrogate(self, x: np.ndarray) -> float:
    if self._restricted_surrogate is None:
      raise ValueError('You must call regress() first.')
    return self._restricted_surrogate(x)


class HarmonicaDesigner(vza.Designer):
  """Harmonica Designer.

  The summary of the current implementation is as follows:

  1. Use previous trials for data collection.
  2. Perform Polynomial Lasso over the data and obtain a predictor function
  based on filtering only the highest coefficients.
  3. Obtain the "J"-set (i.e. set of input indices that only affect the
  predictor function).
  4. Find the global optimizer over the J-set domain on the predictor function,
  and return it as a suggestion.
  """

  def __init__(self,
               problem_statement: vz.ProblemStatement,
               harmonica_q: Optional[HarmonicaQ] = None,
               acquisition_samples: int = 100,
               num_init_samples: int = 10):
    """Init.

    Args:
      problem_statement: Must use a boolean search space.
      harmonica_q: HarmonicaQ class. If None, will use default class with
        default kwargs.
      acquisition_samples: Number of trial samples to optimize final acquisition
        function.
      num_init_samples: Number of initial random suggestions for seeding the
        model.
    """

    if problem_statement.search_space.is_conditional:
      raise ValueError(
          f'This designer {self} does not support conditional search.')
    for p_config in problem_statement.search_space.parameters:
      if p_config.external_type != vz.ExternalType.BOOLEAN:
        raise ValueError('Only boolean search spaces are supported.')

    self._problem_statement = problem_statement
    self._metric_name = self._problem_statement.metric_information.item().name
    self._search_space = problem_statement.search_space
    self._num_vars = len(self._search_space.parameters)

    self._harmonica_q = harmonica_q or HarmonicaQ()
    self._acquisition_samples = acquisition_samples
    self._num_init_samples = num_init_samples

    self._trials = []

  def update(self, trials: vza.CompletedTrials) -> None:
    self._trials += tuple(trials.completed)

  def suggest(self,
              count: Optional[int] = None) -> Sequence[vz.TrialSuggestion]:
    """Performs entire q-stage Harmonica using previous trials for regression data.

    Args:
      count: Makes best effort to generate this many suggestions. If None,
        suggests as many as the algorithm wants.

    Returns:
      New suggestions.
    """
    count = count or 1
    if count > 1:
      raise ValueError('This designer does not support batched suggestions.')

    if len(self._trials) < self._num_init_samples:
      random_designer = random.RandomDesigner(self._search_space)
      return random_designer.suggest(count)

    # Convert previous trial data into regression data.
    X = []
    Y = []
    for t in self._trials:
      single_x = [
          1.0 if t.parameters.get_value(p.name) == 'True' else -1.0
          for p in self._search_space.parameters
      ]
      single_y = t.final_measurement.metrics[self._metric_name].value
      X.append(single_x)
      Y.append(single_y)
    X = np.array(X)
    Y = np.array(Y)

    if self._problem_statement.metric_information.item(
    ).goal == vz.ObjectiveMetricGoal.MINIMIZE:
      Y = -Y

    # Perform q-stage Harmonica.
    self._harmonica_q.reset()
    self._harmonica_q.regress(X, Y)

    # Optimize final acquisition function.
    # TODO: Allow any designer instead of just random search.
    X_temp = np.random.choice([-1.0, 1.0],
                              size=(self._acquisition_samples, self._num_vars))
    Y_temp = np.array([self._harmonica_q.surrogate(x) for x in X_temp])
    x_new = X_temp[np.argmax(Y_temp)]

    parameters = vz.ParameterDict()
    for i, p in enumerate(self._search_space.parameters):
      parameters[p.name] = 'True' if x_new[i] == 1.0 else 'False'
    return [vz.TrialSuggestion(parameters=parameters)]



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Wrappers for Designer into Policy."""
import abc
import json
from typing import Callable, Generic, Sequence, Type, TypeVar, Protocol

from absl import logging
from vizier import algorithms as vza
from vizier import pythia
from vizier import pyvizier as vz
from vizier.interfaces import serializable

_T = TypeVar('_T')


class DesignerFactory(Protocol[_T]):
  """Protocol (PEP-544) for a designer factory."""

  def __call__(self, problem: vz.ProblemStatement) -> _T:
    pass


class DesignerPolicy(pythia.Policy):
  """Wraps a Designer into a pythia Policy.

  > IMPORTANT: If your Designer class is (partially) serializable, use
  > (Partially)SerializableDesignerPolicy instead.

  When a Designer cannot be (partially) serialized, we create a new Designer
  instance at the start of each `suggest()` call. The most recently used
  Designer instance can be saved in self._designer. However, it is for
  interactive analysis/debugging purposes only and never used in
  future `suggest()` calls.
  """

  def __init__(
      self,
      supporter: pythia.PolicySupporter,
      designer_factory: DesignerFactory[vza.Designer],
  ):
    """Init.

    Args:
      supporter:
      designer_factory:
    """
    self._supporter = supporter
    self._designer_factory = designer_factory

  def suggest(self, request: pythia.SuggestRequest) -> pythia.SuggestDecision:
    designer = self._designer_factory(request.study_config)
    new_trials = self._supporter.GetTrials(
        status_matches=vz.TrialStatus.COMPLETED)
    designer.update(vza.CompletedTrials(new_trials))
    self._designer = designer  # saved for debugging purposes only.
    return pythia.SuggestDecision(
        designer.suggest(request.count), metadata=vz.MetadataDelta())

  def early_stop(self,
                 request: pythia.EarlyStopRequest) -> pythia.EarlyStopDecisions:
    raise NotImplementedError(
        'DesignerPolicy does not support the early_stop() method.')


class _SerializableDesignerPolicyBase(pythia.Policy,
                                      serializable.PartiallySerializable,
                                      Generic[_T], abc.ABC):
  """Partially implemented class for wrapping a (Partially)SerializableDesigner.

  Inherited by (Partially)SerializableDesignerPolicy which is fully implemented.

  (Partially)SerializableDesignerPolicy maintains a synchronized state between
  the set of trial ids that were passed to Designer via `update()` call, and
  the Designer instance that was used for during last `suggest()` call.

  The policy's `dump()` contains the trial ids passed to update() and the
  result of `dump()` called on the wrapped Designer.

  Unlike in the basic DesignerPolicy class, this class tries to minimize
  the computation by following these steps in order:
    * Re-use the saved Designer instance from the last `suggest()` call.
    * `load()` the Designer state from the study-level metadata.
    * If either of the first two steps succeedes, then we update the Designer
    with newly completed trials only.
    * Otherwise, we update the Designer with all trials.

  > NOTE: This Policy itself is PartiallySerializable.
  """

  _ns_designer = 'designer'

  def __init__(
      self,
      problem_statement: vz.ProblemStatement,
      supporter: pythia.PolicySupporter,
      designer_factory: Callable[[vz.ProblemStatement], _T],
      *,
      ns_root: str = 'designer_policy_v0',
      verbose: int = 0,
  ):
    """Init.

    Args:
      problem_statement:
      supporter:
      designer_factory:
      ns_root: Root of the namespace where policy state is stored.
      verbose: Logging verbosity.
    """
    self._supporter = supporter
    self._designer_factory = designer_factory
    self._ns_root = ns_root
    self._incorporated_trial_ids = set()
    self._problem_statement = problem_statement
    self._verbose = verbose
    self._designer = None

  def suggest(self, request: pythia.SuggestRequest) -> pythia.SuggestDecision:
    """Perform a suggest operation.

    The order of operations is:
    1. Initialize the designer and load its state from metadata.
    2. Update the designer with newly completed trials.
    3. Generate suggestions from the designer.
    4. Dump the state of the designer and store it in metadata.

    Arguments:
      request: Pythia suggestion request objects.

    Returns:
      The suggestions from the designer.
    """
    # Note that we can avoid O(Num trials) dependency in the standard scenario,
    # by storing only the last element in a consecutive sequence, e.g.,
    # instead of storing [1,2,3,4,11,12,13,21], store: [4,13,21], but
    # we keep things simple in this pseudocode.
    self._initialize_designer(request.study_config)
    new_trials = self._get_new_trials(request.max_trial_id)
    self.designer.update(vza.CompletedTrials(new_trials))
    self._incorporated_trial_ids |= set(t.id for t in new_trials)

    logging.info(
        'Updated with %s trials. Designer has seen a total of %s trials.',
        len(new_trials),
        len(self._incorporated_trial_ids),
    )
    metadata_delta = vz.MetadataDelta()
    # During the 'suggest' call the designer's state could be changed, therefore
    # the state is dumped and stored only after 'suggest' was called.
    suggestions = pythia.SuggestDecision(
        self.designer.suggest(request.count), metadata=metadata_delta)
    metadata_delta.on_study.ns(self._ns_root).attach(self.dump())
    return suggestions

  def early_stop(self,
                 request: pythia.EarlyStopRequest) -> pythia.EarlyStopDecisions:
    raise NotImplementedError(
        'PartiallySerializableDesignerPolicy does not implement early_stop().')

  @property
  def designer(self) -> _T:
    if self._designer is None:
      raise ValueError('`self._designer` has not been initialized!'
                       'Use self._restore_designer(..) to initialize it.')
    return self._designer

  @abc.abstractmethod
  def _restore_designer(self, designer_metadata: vz.Metadata) -> _T:
    """Creates a new Designer by restoring the state from `designer_metadata`.

    Args:
      designer_metadata:

    Returns:
      New Designer object.

    Raises:
      DecodeError: `designer_metadata` does not contain valid information
        to restore a Designer state.
    """

  # TODO: Use timestamps to avoid metadata blowup.
  def load(self, md: vz.Metadata) -> None:
    if 'incorporated_trial_ids' in md:
      try:
        self._incorporated_trial_ids = set(
            json.loads(md['incorporated_trial_ids']))
      except json.JSONDecodeError as e:
        raise serializable.HarmlessDecodeError from e
    else:
      raise serializable.HarmlessDecodeError()
    logging.info(
        'Successfully recovered the policy state, which incorporated %s trials',
        len(self._incorporated_trial_ids),
    )
    self._designer = self._restore_designer(md.ns(self._ns_designer))

  def _initialize_designer(self,
                           problem_statement: vz.ProblemStatement) -> None:
    """Guarantees that `self._designer` is populated.

    This method guarantees that after a successful call, `self.designer` does
    not raise an Exception.
    This method catches all DecodeErrors.

    Args:
      problem_statement:

    Raises:
      ValueError: If problem_statement is differerent from the initially
        received problem_statement.
    """
    if self._designer is not None:
      # When the same policy object is maintained in RAM, prefer keeping
      # the designer object over restoring the state from metadata.
      # TOCONSIDER: Adding a boolean knob to turn off this behavior.
      logging.log_if(
          logging.INFO,
          ('Policy already has a designer. '
           'It will not attempt to load from metadata.'),
          self._verbose >= 2,
      )
      return
    elif self._problem_statement != problem_statement:
      raise ValueError(
          f'{type(self)} cannot be re-used for different study configs!'
          f'Policy: {self}, previous study: {self._problem_statement} '
          f'new study: {problem_statement}')

    metadata = problem_statement.metadata.ns(self._ns_root)
    try:
      self.load(metadata)
      logging.log_if(logging.INFO, 'Successfully decoded all states!',
                     self._verbose >= 1)
    except serializable.DecodeError as e:
      logging.log_if(logging.INFO, 'Failed to decode state. %s',
                     self._verbose >= 1, e)
      self._designer = self._designer_factory(problem_statement)
      self._incorporated_trial_ids = set()

  def dump(self) -> vz.Metadata:
    """Dump state.

    Returns:
      Metadata has the following namespace hierarchy:
        Namespace([self._ns_root]): contains the policy's state.
        Namespace([self._ns_root, self._ns_designer]: contains the designer's
          state.
    """
    md = vz.Metadata()
    md.ns(self._ns_designer).attach(self.designer.dump())
    # TODO: Storing every id is inefficient. Optimize this.
    md['incorporated_trial_ids'] = json.dumps(
        list(self._incorporated_trial_ids))
    return md

  def _get_new_trials(self, max_trial_id: int) -> Sequence[vz.CompletedTrial]:
    """Returns new completed trials that designer should be updated with."""
    if len(self._incorporated_trial_ids) == max_trial_id:
      # no trials need to be loaded.
      return []
    all_trial_ids = set(range(1, max_trial_id + 1))
    trial_ids_to_load = all_trial_ids - self._incorporated_trial_ids

    trials = self._supporter.GetTrials(
        trial_ids=trial_ids_to_load, status_matches=vz.TrialStatus.COMPLETED)
    logging.info(
        ('Loaded %s completed trials out of %s total unseen trials. '
         'Max trial id is %s.'),
        len(trials),
        len(trial_ids_to_load),
        max_trial_id,
    )
    return trials


class PartiallySerializableDesignerPolicy(
    _SerializableDesignerPolicyBase[vza.PartiallySerializableDesigner]):
  """Wraps a PartiallySerializableDesigner."""

  def _restore_designer(
      self,
      designer_metadata: vz.Metadata) -> vza.PartiallySerializableDesigner:
    designer = self._designer_factory(self._problem_statement)
    designer.load(designer_metadata)
    return designer


class SerializableDesignerPolicy(
    _SerializableDesignerPolicyBase[vza.SerializableDesigner]):
  """Wraps a SerializableDesigner."""

  def __init__(
      self,
      problem_statement: vz.ProblemStatement,
      supporter: pythia.PolicySupporter,
      designer_factory: DesignerFactory[vza.SerializableDesigner],
      designer_cls: Type[vza.SerializableDesigner],
      *,
      ns_root: str = 'designer_policy_v0',
      verbose: int = 0,
  ):
    """Init.

    Args:
      problem_statement:
      supporter:
      designer_factory: Used when designer state cannot be restored.
      designer_cls: Type name of the designer. Its load() classmethod is called
        to restore the designer state.
      ns_root: Root of the namespace where policy state is stored.
      verbose: Logging verbosity.
    """
    super().__init__(
        problem_statement,
        supporter,
        designer_factory,
        ns_root=ns_root,
        verbose=verbose,
    )
    self._designer_cls = designer_cls

  def _restore_designer(
      self, designer_metadata: vz.Metadata) -> vza.SerializableDesigner:
    return self._designer_cls.recover(designer_metadata)



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Coroutine for a trainable Gaussian process with an ARD kernel."""

from typing import Any, Generator, Optional, Type, Union

from jax import numpy as jnp
from tensorflow_probability.substrates import jax as tfp
from vizier._src.jax import stochastic_process_model as sp_model

Array = Any
tfd = tfp.distributions
tfpk = tfp.math.psd_kernels


class GaussianProcessARD(sp_model.ModelCoroutine):
  """Specifies a Gaussian process model with an ARD kernel.

  `GaussianProcessARD` satisfies the `ModelCoroutine` Protocol. By default,
  `GaussianProcessARD` uses a Matern Five Halves kernel. Hyperparameter priors
  are LogNormal(0., 1.).

  #### Examples

  Build a Gaussian Process model with an ARD Exponentiated Quadratic kernel and
  LogNormal(0., 1.) hyperparameter priors for 5-dimensional data.

  ```python
  from tensorflow_probability.substrates import jax as tfp

  tfpk = tfp.math.psd_kernels

  dim = 5
  gen = GaussianProcessARD(
      dimension=dim,
      kernel_class=tfpk.ExponentiatedQuadratic)
  ```
  """

  def __init__(self,
               dimension: int,
               kernel_class: Union[
                   Type[tfpk.MaternFiveHalves], Type[tfpk.MaternThreeHalves],
                   Type[tfpk.MaternOneHalf], Type[tfpk.Parabolic],
                   Type[tfpk.ExponentiatedQuadratic]] = tfpk.MaternFiveHalves,
               *,
               use_tfp_runtime_validation: bool = False):
    """Initializes a `GaussianProcessARD`.

    Args:
      dimension: The data dimensionality.
      kernel_class: The Gaussian process' kernel type.
      use_tfp_runtime_validation: If True, run additional runtime checks on the
        validity of the parameters and input for TFP objects (e.g. verify that
        amplitude and length scale parameters are positive). Runtime checks may
        be expensive and should be used only during development/debugging.
    """
    self.dimension = dimension
    self._kernel_class = kernel_class
    self._use_tfp_runtime_validation = use_tfp_runtime_validation

  def __call__(
      self, inputs: Optional[Array] = None
  ) -> Generator[sp_model.ModelParameter, Array, tfd.GaussianProcess]:
    # TODO: Determine why pylint doesn't allow both Returns and
    # Yields sections.
    # pylint: disable=g-doc-return-or-yield
    """The coroutine that specifies the GP model.

    Args:
      inputs: index_points to be provided to the GP.

    Yields:
      `ModelParameter`s describing the parameters to be declared in the Flax
        model.

    Returns:
      A tfd.GaussianProcess with the given index points.
    """
    amplitude = yield sp_model.ModelParameter.from_prior(
        tfd.LogNormal(0.0, 1.0, name='amplitude'),
        constraint=sp_model.Constraint(bounds=(jnp.array(0.0), None)),
    )
    kernel = self._kernel_class(
        amplitude=amplitude,
        length_scale=1.,
        validate_args=self._use_tfp_runtime_validation)
    inverse_length_scale = yield sp_model.ModelParameter.from_prior(
        tfd.Sample(
            tfd.LogNormal(0.0, 1.0),
            sample_shape=(self.dimension,),
            name='inverse_length_scale',
        ),
        constraint=sp_model.Constraint(bounds=(jnp.array(0.0), None)),
    )
    kernel = tfpk.FeatureScaled(
        kernel,
        inverse_scale_diag=inverse_length_scale,
        validate_args=self._use_tfp_runtime_validation)
    observation_noise_variance = yield sp_model.ModelParameter.from_prior(
        tfd.LogNormal(0.0, 1.0, name='observation_noise_variance'),
        constraint=sp_model.Constraint(bounds=(jnp.array(0.0), None)),
    )
    return tfd.GaussianProcess(
        kernel,
        index_points=inputs,
        observation_noise_variance=observation_noise_variance,
        validate_args=self._use_tfp_runtime_validation,
        always_yield_multivariate_normal=True)



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Abstractions."""

import abc
from typing import Optional, Sequence, TypeVar

import attr
from vizier import pyvizier as vz
from vizier.interfaces import serializable

_T = TypeVar('_T')


@attr.define(frozen=True)
class CompletedTrials:
  """A group of completed trials.

  Attributes:
    completed: Completed Trials.
  """

  def __attrs_post_init__(self):
    for trial in self.completed:
      if trial.status != vz.TrialStatus.COMPLETED:
        raise ValueError(f'All trials must be completed. Bad trial:\n{trial}')

  completed: Sequence[vz.Trial] = attr.field(
      converter=tuple,
      validator=attr.validators.deep_iterable(
          attr.validators.instance_of(vz.Trial)))


class _SuggestionAlgorithm(abc.ABC):
  """Suggestion algorithm."""

  @abc.abstractmethod
  def suggest(self,
              count: Optional[int] = None) -> Sequence[vz.TrialSuggestion]:
    """Make new suggestions.

    Args:
      count: Makes best effort to generate this many suggestions. If None,
        suggests as many as the algorithm wants.

    Returns:
      New suggestions.
    """
    pass


class Designer(_SuggestionAlgorithm):
  """Suggestion algorithm for sequential usage.

  Designer is the recommended interface for implementing commonly used
  algorithms such as GP-UCB and evolutionary algorithms. A Designer can be
  wrapped into a pythia `Policy` via `DesignerPolicy`. When run inside a service
  binary, a Designer instance does not persist during the lifetime of a `Study`.
  It receives all trials from the beginning of the study in `update()` calls.
  This can be inefficient.

  If your Designer can benefit from a persistent state, implement
  `(Partially)SerializableDesigner` interface and use
  `(Partially)SerializableDesignerPolicy` to wrap it.
  Vizier service will serialize the Designer's state in DB, restore it for
  the next usage, and update it with the newly completed trials since the last
  suggestion.

  > IMPORTANT: If your Designer changes its state inside `suggest()` (e.g. to
  > incorporate its own suggestions before completion), then use
  > (Partially)SerializableDesigner interface instead.
  """

  @abc.abstractmethod
  def update(self, delta: CompletedTrials) -> None:
    """Incorporates newly completed trials into the designer's state."""
    pass


class PartiallySerializableDesigner(Designer,
                                    serializable.PartiallySerializable):
  pass


class SerializableDesigner(Designer, serializable.Serializable):
  pass



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for outputwarpers."""

import numpy as np
import scipy
from vizier._src.algorithms.designers.gp import output_warpers

from absl.testing import absltest
from absl.testing import parameterized

OutputWarper = output_warpers.OutputWarperProtocol


@absltest.skipThisClass('Base class')
class _OutputWarperTestCase(absltest.TestCase):

  @property
  def warper(self) -> OutputWarper:
    raise RuntimeError('Subclasses should override this method!')

  @property
  def always_maps_to_finite(self) -> bool:
    # Override it to True if the warper should map every value to a
    # finite value.
    return False

  def labels_with_outliers(self):
    return np.array([[1.], [1.], [5.], [-1e80], [np.nan], [-np.inf]])

  def test_always_maps_to_finite(self):
    if not self.always_maps_to_finite:
      self.skipTest('This class does not map every value to a finite value.')

    labels = np.array([[1.], [1.], [5.], [-1e80], [np.nan], [-np.inf]])
    labels_warped = self.warper(labels)
    np.testing.assert_allclose(
        np.isfinite(labels_warped), True, err_msg=f'warped: {labels_warped}')

  def test_input_is_not_mutated(self):
    labels_input = np.array([[1.], [1.], [5.], [10.]])
    _ = self.warper(labels_input)
    self.assertTrue(
        (labels_input.flatten() == np.array([[1.], [1.], [5.],
                                             [10.]]).flatten()).all())

  def test_shape_is_preserved(self):
    labels = self.labels_with_outliers()
    labels_warped = self.warper(labels)
    self.assertEqual(labels_warped.shape, labels.shape)

  def test_preserve_rank_despite_outliers(self):
    labels = self.labels_with_outliers()
    finite_indices = np.isfinite(labels)
    labels_warped = self.warper(labels)
    np.testing.assert_array_equal(
        scipy.stats.rankdata(labels[finite_indices]),
        scipy.stats.rankdata(labels_warped[finite_indices]),
        err_msg=(f'Unwarped: {labels}\n'
                 f'Warped: {labels_warped}'))

  def test_preserve_rank_if_no_outliers(self):
    labels = np.array([[1.], [1.], [5.], [-1], [-4], [np.nan], [-np.inf]])
    finite_indices = np.isfinite(labels)
    labels_warped = self.warper(labels)
    np.testing.assert_array_equal(
        scipy.stats.rankdata(labels[finite_indices]),
        scipy.stats.rankdata(labels_warped[finite_indices]),
        err_msg=(f'Unwarped: {labels}\n'
                 f'Warped: {labels_warped}'))

  def test_finite_maps_to_finite(self):
    labels = self.labels_with_outliers()
    finite_indices = np.isfinite(labels)
    labels_warped = self.warper(labels)
    np.testing.assert_allclose(
        np.isfinite(labels_warped[finite_indices]),
        True,
        err_msg=f'warped: {labels_warped}')


class DefaultOutputWarperTest(_OutputWarperTestCase, parameterized.TestCase):

  @property
  def warper(self) -> OutputWarper:
    return output_warpers.create_default_warper(infeasible_warp=True)

  @property
  def always_maps_to_finite(self) -> bool:
    return True

  def test_all_nonfinite_labels(self):
    labels_infeaible = np.array([[-np.inf], [np.nan], [np.nan], [-np.inf]])
    self.assertTrue((self.warper(labels_infeaible) == -1 *
                     np.ones(shape=labels_infeaible.shape).flatten()).all())

  @parameterized.parameters([
      dict(labels=np.zeros(shape=(5, 1))),
      dict(labels=np.ones(shape=(5, 1))),
      dict(labels=100 * np.ones(shape=(5, 1))),
      dict(labels=-100 * np.ones(shape=(5, 1))),
  ])
  def test_all_identical_values_map_to_zero(self, labels):
    np.testing.assert_array_equal(self.warper(labels), 0.)

  @parameterized.named_parameters([
      dict(
          testcase_name='case1',
          unwarped=np.array([[1.], [1.], [5.], [-1e80], [np.nan], [-np.inf]]),
          expected=np.array([[-0.14118114], [-0.14118114], [0.5], [-0.5], [-2.],
                             [-2.]]),
      ),
  ])
  def test_known_arrays(self, unwarped: np.ndarray, expected: np.ndarray):
    actual = self.warper(unwarped)
    np.testing.assert_allclose(actual, expected, err_msg=f'actual: {actual}')

  def test_default_warper_empty_warpers(self):
    with self.assertRaises(ValueError):
      output_warpers.create_default_warper(
          half_rank_warp=False, log_warp=False, infeasible_warp=False)


class ZScoreLabelsTest(_OutputWarperTestCase):

  @property
  def warper(self) -> OutputWarper:
    return output_warpers.ZScoreLabels()

  def test_preserve_rank_despite_outliers(self):
    # TODO: Fix this test, or add an explanation why this can be skipped.
    pass

  def test_known_arrays(self):
    # TODO: Add a couple of parameterized test cases.
    self.skipTest('No test cases provided')


class NormalizeLabelsTest(_OutputWarperTestCase):

  @property
  def warper(self) -> OutputWarper:
    return output_warpers.NormalizeLabels()

  def test_known_arrays(self):
    # TODO: Add a couple of parameterized test cases.
    self.skipTest('No test cases provided')


class DetectOutliersTest(_OutputWarperTestCase):

  @property
  def warper(self) -> OutputWarper:
    return output_warpers.create_warp_outliers_warper()

  # TODO: Add extra test coverage for the warp_outliers_warper.

  @property
  def always_maps_to_finite(self) -> bool:
    return True


class TransformToGaussianTest(_OutputWarperTestCase):

  @property
  def warper(self) -> OutputWarper:
    return output_warpers.TransformToGaussian()

  def test_finite_maps_to_finite(self):
    # TODO: Fix this test, or add an explanation why this can be skipped.
    pass

  def test_preserve_rank_if_no_outliers(self):
    # TODO: Fix this test, or add an explanation why this can be skipped.
    pass

  def test_preserve_rank_despite_outliers(self):
    # TODO: Fix this test, or add an explanation why this can be skipped.
    pass


class HalfRankComponentTest(_OutputWarperTestCase, parameterized.TestCase):

  @property
  def warper(self) -> output_warpers.HalfRankComponent:
    return output_warpers.HalfRankComponent()

  @parameterized.named_parameters([
      dict(
          testcase_name='case1',
          unwarped=np.array([[np.nan, 1, 2, 3, 4, 5, 6, 2, 10, 12, -np.inf]]).T,
          expected=np.array([[np.nan], [-2.7145447657886415],
                             [0.3722561569665319], [2.322289907556879], [4.0],
                             [5.0], [6.0], [0.3722561569665319], [10.0], [12.0],
                             [np.nan]])),
      dict(
          testcase_name='case2',
          unwarped=np.array([[np.nan, -4, -3, -2, 1.1, 1.2, 1.3, 1.4, 1.5]]).T,
          expected=np.array([[np.nan], [0.7984888240158797],
                             [0.9467291870388195], [1.0380072549079085],
                             [1.1139555940074284], [1.2], [1.3], [1.4], [1.5]]),
      ),
      dict(
          testcase_name='case3',
          unwarped=np.array([[np.nan, 1, 2, 3, 4, 4, 6, 7, 10, 11, 12]],
                            dtype=np.float64).T,
          expected=np.array([[np.nan], [-2.3573836671676096],
                             [0.7453945664588675], [2.655910679724611],
                             [4.2455644597926385], [4.2455644597926385], [6.0],
                             [7.0], [10.0], [11.0], [12.0]]),
      ),
  ])
  def test_known_arrays(self, unwarped: np.ndarray, expected: np.ndarray):
    actual = self.warper(unwarped)
    np.testing.assert_allclose(
        actual, expected, err_msg=f'actual: {actual.tolist()}')


class LogWarperComponentTest(_OutputWarperTestCase):

  @property
  def warper(self) -> OutputWarper:
    return output_warpers.LogWarperComponent()

  def test_preserve_rank_despite_outliers(self):
    # TODO: Fix this test, or add an explanation why this can be skipped.
    pass

  def test_known_arrays(self):
    # TODO: Add a couple of parameterized test cases.
    self.skipTest('No test cases provided')


class InfeasibleWarperTest(parameterized.TestCase):

  @property
  def always_maps_to_finite(self) -> bool:
    return True

  def test_warper_removes_nans(self):
    warper_infeasible = output_warpers.InfeasibleWarperComponent()
    labels = np.array([[-200.], [np.nan], [-1000.], [np.nan], [1.], [2.], [3.]])

    labels_warped_infeasible = warper_infeasible(labels)
    self.assertEqual(np.isnan(labels_warped_infeasible).sum(), 0)

  def test_known_arrays(self):
    # TODO: Add a couple of parameterized test cases.
    self.skipTest('No test cases provided')

if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for clients."""

from absl import flags
from absl import logging
from vizier.client import client_abc_testing
from vizier.service import clients
from vizier.service import pyvizier as vz
from vizier.service import vizier_server

from absl.testing import absltest

FLAGS = flags.FLAGS


class VizierClientTest(client_abc_testing.TestCase):
  _owner: str

  def create_study(
      self, problem: vz.ProblemStatement, study_id: str
  ) -> clients.Study:
    config = vz.StudyConfig.from_problem(problem)
    config.algorithm = vz.Algorithm.RANDOM_SEARCH
    study = clients.Study.from_study_config(
        config, owner='owner', study_id=study_id
    )
    return study

  def create_study2(
      self, problem: vz.ProblemStatement, study_id: str
  ) -> clients.Study:
    config = vz.StudyConfig.from_problem(problem)
    config.algorithm = vz.Algorithm.RANDOM_SEARCH
    study = clients.Study.from_owner_and_id(owner='owner', study_id=study_id)
    return study

  def test_e2e_tuning(self):
    self.assertPassesE2ETuning()


class VizierClientTestOnServicer(VizierClientTest):

  @classmethod
  def setUpClass(cls):
    logging.info('Test setup started.')
    super().setUpClass()
    clients.environment_variables.server_endpoint = clients.NO_ENDPOINT
    logging.info('Test setup finished.')


class VizierClientTestOnDefaultServer(VizierClientTest):
  _server: vizier_server.DefaultVizierServer

  @classmethod
  def setUpClass(cls):
    logging.info('Test setup started.')
    super().setUpClass()
    cls._server = vizier_server.DefaultVizierServer()
    clients.environment_variables.server_endpoint = cls._server.endpoint
    logging.info('Test setup finished.')

  @classmethod
  def tearDownClass(cls):
    cls._server._server.stop(None)
    super().tearDownClass()


class VizierClientTestOnDistributedPythiaServer(VizierClientTest):
  _server: vizier_server.DistributedPythiaVizierServer

  @classmethod
  def setUpClass(cls):
    logging.info('Test setup started.')
    super().setUpClass()
    cls._server = vizier_server.DistributedPythiaVizierServer()
    clients.environment_variables.server_endpoint = cls._server.endpoint
    logging.info('Test setup finished.')

  @classmethod
  def tearDownClass(cls):
    cls._server._server.stop(None)
    cls._server._pythia_server.stop(None)
    super().tearDownClass()


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for acquisitions."""
import mock
import numpy as np
from vizier._src.algorithms.designers.gp import acquisitions
from vizier.pyvizier import converters

from absl.testing import absltest


def _build_mock_continuous_array_specs(n):
  continuous_spec = mock.create_autospec(converters.NumpyArraySpec)
  continuous_spec.type = converters.NumpyArraySpecType.CONTINUOUS
  continuous_spec.num_dimensions = 1
  return [continuous_spec] * n


class TrustRegionTest(absltest.TestCase):

  def test_trust_region_small(self):
    tr = acquisitions.TrustRegion(
        np.array([
            [0.0, 0.0, 0.0, 0.0],
            [1.0, 1.0, 1.0, 1.0],
        ]),
        _build_mock_continuous_array_specs(4),
    )

    np.testing.assert_allclose(
        tr.min_linf_distance(
            np.array([
                [0., .2, .3, 0.],
                [.9, .8, .9, .9],
                [1., 1., 1., 1.],
            ]),), np.array([0.3, 0.2, 0.]))
    self.assertAlmostEqual(tr.trust_radius, 0.224, places=3)

  def test_trust_region_bigger(self):
    tr = acquisitions.TrustRegion(
        np.vstack(
            [
                [0.0, 0.0, 0.0, 0.0],
                [1.0, 1.0, 1.0, 1.0],
            ]
            * 10
        ),
        _build_mock_continuous_array_specs(4),
    )
    np.testing.assert_allclose(
        tr.min_linf_distance(
            np.array([
                [0., .2, .3, 0.],
                [.9, .8, .9, .9],
                [1., 1., 1., 1.],
            ]),), np.array([0.3, 0.2, 0.]))
    self.assertAlmostEqual(tr.trust_radius, 0.44, places=3)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Implementation of SQL Datastore."""
import collections
import threading
from typing import Callable, DefaultDict, Iterable, List, Optional
from absl import logging

import sqlalchemy as sqla

from vizier.service import custom_errors
from vizier.service import datastore
from vizier.service import key_value_pb2
from vizier.service import resources
from vizier.service import study_pb2
from vizier.service import vizier_oss_pb2
from google.longrunning import operations_pb2


# TODO: Consider using ORM API (when fixed) to reduce code length.
class SQLDataStore(datastore.DataStore):
  """SQL Datastore."""

  def __init__(self, engine):
    self._engine = engine
    self._connection = self._engine.connect()
    self._root_metadata = sqla.MetaData()
    self._owners_table = sqla.Table(
        'owners',
        self._root_metadata,
        sqla.Column('owner_name', sqla.String, primary_key=True),
    )
    self._studies_table = sqla.Table(
        'studies',
        self._root_metadata,
        sqla.Column('study_name', sqla.String, primary_key=True),
        sqla.Column('owner_id', sqla.String),
        sqla.Column('study_id', sqla.String),
        sqla.Column('serialized_study', sqla.String),
    )
    self._trials_table = sqla.Table(
        'trials',
        self._root_metadata,
        sqla.Column('trial_name', sqla.String, primary_key=True),
        sqla.Column('owner_id', sqla.String),
        sqla.Column('study_id', sqla.String),
        sqla.Column('trial_id', sqla.INTEGER),
        sqla.Column('serialized_trial', sqla.String),
    )
    self._suggestion_operations_table = sqla.Table(
        'suggestion_operations',
        self._root_metadata,
        sqla.Column('operation_name', sqla.String, primary_key=True),
        sqla.Column('owner_id', sqla.String),
        sqla.Column('study_id', sqla.String),
        sqla.Column('client_id', sqla.String),
        sqla.Column('operation_number', sqla.INTEGER),
        sqla.Column('serialized_op', sqla.String),
    )
    self._early_stopping_operations_table = sqla.Table(
        'early_stopping_operations',
        self._root_metadata,
        sqla.Column('operation_name', sqla.String, primary_key=True),
        sqla.Column('owner_id', sqla.String),
        sqla.Column('study_id', sqla.String),
        sqla.Column('trial_id', sqla.INTEGER),
        sqla.Column('serialized_op', sqla.String),
    )
    # This lock is meant to lock `execute()` calls for database types which
    # don't support multi-threading, like SQLite.
    self._lock = threading.Lock()
    self._root_metadata.create_all(self._engine)

  def create_study(self, study: study_pb2.Study) -> resources.StudyResource:
    study_resource = resources.StudyResource.from_name(study.name)
    owner_name = study_resource.owner_resource.name
    owner_query = self._owners_table.insert().values(owner_name=owner_name)
    study_query = self._studies_table.insert().values(
        study_name=study.name,
        owner_id=study_resource.owner_id,
        study_id=study_resource.study_id,
        serialized_study=study.SerializeToString(),
    )

    with self._lock:
      try:
        self._connection.execute(owner_query)
      except sqla.exc.IntegrityError:
        logging.info('Owner with name %s currently exists.', owner_name)
      try:
        self._connection.execute(study_query)
        return study_resource
      except sqla.exc.IntegrityError as integrity_error:
        raise custom_errors.AlreadyExistsError(
            'Study with name %s already exists.' % study.name
        ) from integrity_error

  def load_study(self, study_name: str) -> study_pb2.Study:
    query = sqla.select([self._studies_table])
    query = query.where(self._studies_table.c.study_name == study_name)

    with self._lock:
      result = self._connection.execute(query)

    row = result.fetchone()
    if not row:
      raise custom_errors.NotFoundError(
          'Failed to find study name: %s' % study_name
      )
    return study_pb2.Study.FromString(row['serialized_study'])

  def update_study(self, study: study_pb2.Study) -> resources.StudyResource:
    study_resource = resources.StudyResource.from_name(study.name)
    exists_query = sqla.exists(
        sqla.select([self._studies_table]).where(
            self._studies_table.c.study_name == study.name
        )
    ).select()
    update_query = (
        sqla.update(self._studies_table)
        .where(self._studies_table.c.study_name == study.name)
        .values(
            study_name=study.name,
            owner_id=study_resource.owner_id,
            study_id=study_resource.study_id,
            serialized_study=study.SerializeToString(),
        )
    )

    with self._lock:
      exists = self._connection.execute(exists_query).fetchone()[0]
      if not exists:
        raise custom_errors.NotFoundError(
            'Study %s does not exist.' % study.name
        )
      self._connection.execute(update_query)
    return study_resource

  def delete_study(self, study_name: str) -> None:
    study_resource = resources.StudyResource.from_name(study_name)

    exists_query = sqla.select([self._studies_table])
    exists_query = exists_query.where(
        self._studies_table.c.study_name == study_name
    )
    exists_query = sqla.exists(exists_query).select()
    delete_study_query = self._studies_table.delete().where(
        self._studies_table.c.study_name == study_name
    )
    delete_trials_query = (
        self._trials_table.delete()
        .where(self._trials_table.c.owner_id == study_resource.owner_id)
        .where(self._trials_table.c.study_id == study_resource.study_id)
    )

    with self._lock:
      exists = self._connection.execute(exists_query).fetchone()[0]
      if not exists:
        raise custom_errors.NotFoundError(
            'Study %s does not exist.' % study_name
        )
      self._connection.execute(delete_study_query)
      self._connection.execute(delete_trials_query)

  def list_studies(self, owner_name: str) -> List[study_pb2.Study]:
    owner_id = resources.OwnerResource.from_name(owner_name).owner_id
    exists_query = sqla.exists(
        sqla.select([self._owners_table]).where(
            self._owners_table.c.owner_name == owner_name
        )
    ).select()
    list_query = sqla.select([self._studies_table]).where(
        self._studies_table.c.owner_id == owner_id
    )

    with self._lock:
      exists = self._connection.execute(exists_query).fetchone()[0]
      if not exists:
        raise custom_errors.NotFoundError(
            'Owner name %s does not exist.' % owner_name
        )
      result = self._connection.execute(list_query).fetchall()

    return [
        study_pb2.Study.FromString(row['serialized_study']) for row in result
    ]

  def create_trial(self, trial: study_pb2.Trial) -> resources.TrialResource:
    trial_resource = resources.TrialResource.from_name(trial.name)
    query = self._trials_table.insert().values(
        trial_name=trial.name,
        owner_id=trial_resource.owner_id,
        study_id=trial_resource.study_id,
        trial_id=trial_resource.trial_id,
        serialized_trial=trial.SerializeToString(),
    )

    with self._lock:
      try:
        self._connection.execute(query)
        return trial_resource
      except sqla.exc.IntegrityError as integrity_error:
        raise custom_errors.AlreadyExistsError(
            'Trial with name %s already exists.' % trial.name
        ) from integrity_error

  def get_trial(self, trial_name: str) -> study_pb2.Trial:
    query = sqla.select([self._trials_table])
    query = query.where(self._trials_table.c.trial_name == trial_name)

    with self._lock:
      result = self._connection.execute(query)

    row = result.fetchone()
    if not row:
      raise custom_errors.NotFoundError(
          'Failed to find trial name: %s' % trial_name
      )
    return study_pb2.Trial.FromString(row['serialized_trial'])

  def update_trial(self, trial: study_pb2.Trial) -> resources.TrialResource:
    trial_resource = resources.TrialResource.from_name(trial.name)
    exists_query = sqla.exists(
        sqla.select([self._trials_table]).where(
            self._trials_table.c.trial_name == trial.name
        )
    ).select()
    update_query = (
        sqla.update(self._trials_table)
        .where(self._trials_table.c.trial_name == trial.name)
        .values(
            trial_name=trial.name,
            owner_id=trial_resource.owner_id,
            study_id=trial_resource.study_id,
            trial_id=trial_resource.trial_id,
            serialized_trial=trial.SerializeToString(),
        )
    )

    with self._lock:
      exists = self._connection.execute(exists_query).fetchone()[0]
      if not exists:
        raise custom_errors.NotFoundError(
            'Trial %s does not exist.' % trial.name
        )
      self._connection.execute(update_query)

    return trial_resource

  def list_trials(self, study_name: str) -> List[study_pb2.Trial]:
    study_resource = resources.StudyResource.from_name(study_name)
    exists_query = sqla.exists(
        sqla.select([self._studies_table]).where(
            self._studies_table.c.study_name == study_name
        )
    ).select()
    list_query = (
        sqla.select([self._trials_table])
        .where(self._trials_table.c.owner_id == study_resource.owner_id)
        .where(self._trials_table.c.study_id == study_resource.study_id)
    )

    with self._lock:
      exists = self._connection.execute(exists_query).fetchone()[0]
      if not exists:
        raise custom_errors.NotFoundError(
            'Study name %s does not exist.' % study_name
        )
      result = self._connection.execute(list_query)

    return [
        study_pb2.Trial.FromString(row['serialized_trial']) for row in result
    ]

  def delete_trial(self, trial_name: str) -> None:
    exists_query = sqla.exists(
        sqla.select([self._trials_table]).where(
            self._trials_table.c.trial_name == trial_name
        )
    ).select()
    delete_query = self._trials_table.delete().where(
        self._trials_table.c.trial_name == trial_name
    )
    with self._lock:
      exists = self._connection.execute(exists_query).fetchone()[0]
      if not exists:
        raise custom_errors.NotFoundError(
            'Trial %s does not exist.' % trial_name
        )
      self._connection.execute(delete_query)

  def max_trial_id(self, study_name: str) -> int:
    study_resource = resources.StudyResource.from_name(study_name)
    exists_query = sqla.exists(
        sqla.select([self._studies_table]).where(
            self._studies_table.c.study_name == study_name
        )
    ).select()
    trial_id_query = (
        sqla.select(
            [sqla.func.max(self._trials_table.c.trial_id, type_=sqla.INT)]
        )
        .where(self._trials_table.c.owner_id == study_resource.owner_id)
        .where(self._trials_table.c.study_id == study_resource.study_id)
    )

    with self._lock:
      exists = self._connection.execute(exists_query).fetchone()[0]
      if not exists:
        raise custom_errors.NotFoundError(
            'Study %s does not exist.' % study_name
        )
      potential_trial_id = self._connection.execute(trial_id_query).fetchone()[
          0
      ]

    if potential_trial_id is None:
      return 0
    return potential_trial_id

  def create_suggestion_operation(
      self, operation: operations_pb2.Operation
  ) -> resources.SuggestionOperationResource:
    resource = resources.SuggestionOperationResource.from_name(operation.name)
    query = self._suggestion_operations_table.insert().values(
        operation_name=operation.name,
        owner_id=resource.owner_id,
        study_id=resource.study_id,
        client_id=resource.client_id,
        operation_number=resource.operation_number,
        serialized_op=operation.SerializeToString(),
    )

    try:
      with self._lock:
        self._connection.execute(query)
      return resource
    except sqla.exc.IntegrityError as integrity_error:
      raise custom_errors.AlreadyExistsError(
          'Suggest Op with name %s already exists.' % operation.name
      ) from integrity_error

  def get_suggestion_operation(
      self, operation_name: str
  ) -> operations_pb2.Operation:
    query = sqla.select([self._suggestion_operations_table]).where(
        self._suggestion_operations_table.c.operation_name == operation_name
    )

    with self._lock:
      result = self._connection.execute(query)

    row = result.fetchone()
    if not row:
      raise custom_errors.NotFoundError(
          'Failed to find suggest op name: %s' % operation_name
      )
    return operations_pb2.Operation.FromString(row['serialized_op'])

  def update_suggestion_operation(
      self, operation: operations_pb2.Operation
  ) -> resources.SuggestionOperationResource:
    resource = resources.SuggestionOperationResource.from_name(operation.name)

    exists_query = sqla.exists(
        sqla.select([self._suggestion_operations_table]).where(
            self._suggestion_operations_table.c.operation_name == operation.name
        )
    ).select()
    update_query = (
        sqla.update(self._suggestion_operations_table)
        .where(
            self._suggestion_operations_table.c.operation_name == operation.name
        )
        .values(
            operation_name=operation.name,
            owner_id=resource.owner_id,
            study_id=resource.study_id,
            client_id=resource.client_id,
            operation_number=resource.operation_number,
            serialized_op=operation.SerializeToString(),
        )
    )

    with self._lock:
      exists = self._connection.execute(exists_query).fetchone()[0]
      if not exists:
        raise custom_errors.NotFoundError(
            'Suggest op %s does not exist.' % operation.name
        )
      self._connection.execute(update_query)
    return resource

  def list_suggestion_operations(
      self,
      study_name: str,
      client_id: str,
      filter_fn: Optional[Callable[[operations_pb2.Operation], bool]] = None,
  ) -> List[operations_pb2.Operation]:
    study_resource = resources.StudyResource.from_name(study_name)
    query = sqla.select([self._suggestion_operations_table])
    query = query.where(
        self._suggestion_operations_table.c.owner_id == study_resource.owner_id
    )
    query = query.where(
        self._suggestion_operations_table.c.study_id == study_resource.study_id
    )
    query = query.where(
        self._suggestion_operations_table.c.client_id == client_id
    )

    exists_query = sqla.exists(query).select()
    with self._lock:
      exists = self._connection.execute(exists_query).fetchone()[0]
      if not exists:
        raise custom_errors.NotFoundError(
            'Could not find (study_name, client_id):',
            (study_resource.name, client_id),
        )
      result = self._connection.execute(query)

    all_ops = [
        operations_pb2.Operation.FromString(row['serialized_op'])
        for row in result
    ]
    if filter_fn is not None:
      output_list = []
      for op in all_ops:
        if filter_fn(op):
          output_list.append(op)
      return output_list
    else:
      return all_ops

  def max_suggestion_operation_number(
      self, study_name: str, client_id: str
  ) -> int:
    resource = resources.StudyResource.from_name(study_name)

    exists_query = sqla.exists(
        sqla.select([self._suggestion_operations_table])
        .where(
            self._suggestion_operations_table.c.owner_id == resource.owner_id
        )
        .where(
            self._suggestion_operations_table.c.study_id == resource.study_id
        )
        .where(self._suggestion_operations_table.c.client_id == client_id)
    ).select()
    max_query = (
        sqla.select(
            [
                sqla.func.max(
                    self._suggestion_operations_table.c.operation_number,
                    type_=sqla.INT,
                )
            ]
        )
        .where(
            self._suggestion_operations_table.c.owner_id == resource.owner_id
        )
        .where(
            self._suggestion_operations_table.c.study_id == resource.study_id
        )
        .where(self._suggestion_operations_table.c.client_id == client_id)
    )

    with self._lock:
      exists = self._connection.execute(exists_query).fetchone()[0]
      if not exists:
        raise custom_errors.NotFoundError(
            'Could not find (study_name, client_id):', (study_name, client_id)
        )
      return self._connection.execute(max_query).fetchone()[0]

  def create_early_stopping_operation(
      self, operation: vizier_oss_pb2.EarlyStoppingOperation
  ) -> resources.EarlyStoppingOperationResource:
    resource = resources.EarlyStoppingOperationResource.from_name(
        operation.name
    )
    query = self._early_stopping_operations_table.insert().values(
        operation_name=operation.name,
        owner_id=resource.owner_id,
        study_id=resource.study_id,
        trial_id=resource.trial_id,
        serialized_op=operation.SerializeToString(),
    )

    try:
      with self._lock:
        self._connection.execute(query)
      return resource
    except sqla.exc.IntegrityError as integrity_error:
      raise custom_errors.AlreadyExistsError(
          'Early stopping op with name %s already exists.' % operation.name
      ) from integrity_error

  def get_early_stopping_operation(
      self, operation_name: str
  ) -> vizier_oss_pb2.EarlyStoppingOperation:
    query = sqla.select([self._early_stopping_operations_table]).where(
        self._early_stopping_operations_table.c.operation_name == operation_name
    )

    with self._lock:
      result = self._connection.execute(query)

    row = result.fetchone()
    if not row:
      raise custom_errors.NotFoundError(
          'Failed to find early stopping op name: %s' % operation_name
      )
    return vizier_oss_pb2.EarlyStoppingOperation.FromString(
        row['serialized_op']
    )

  def update_early_stopping_operation(
      self, operation: vizier_oss_pb2.EarlyStoppingOperation
  ) -> resources.EarlyStoppingOperationResource:
    resource = resources.EarlyStoppingOperationResource.from_name(
        operation.name
    )
    exists_query = sqla.exists(
        sqla.select([self._early_stopping_operations_table]).where(
            self._early_stopping_operations_table.c.operation_name
            == operation.name
        )
    ).select()
    update_query = (
        sqla.update(self._early_stopping_operations_table)
        .where(
            self._early_stopping_operations_table.c.operation_name
            == operation.name
        )
        .values(
            operation_name=operation.name,
            owner_id=resource.owner_id,
            study_id=resource.study_id,
            trial_id=resource.trial_id,
            serialized_op=operation.SerializeToString(),
        )
    )

    with self._lock:
      exists = self._connection.execute(exists_query).fetchone()[0]
      if not exists:
        raise custom_errors.NotFoundError(
            'Early stopping op %s does not exist.' % operation.name
        )
      self._connection.execute(update_query)
      return resource

  def update_metadata(
      self,
      study_name: str,
      study_metadata: Iterable[key_value_pb2.KeyValue],
      trial_metadata: Iterable[datastore.UnitMetadataUpdate],
  ) -> None:
    """Store the supplied metadata into the SQL database."""
    s_resource = resources.StudyResource.from_name(study_name)
    logging.debug('database.update_metadata s_resource= %s', s_resource)
    # Obtain original study.
    get_study_query = sqla.select([self._studies_table]).where(
        self._studies_table.c.study_name == study_name
    )

    with self._lock:
      study_result = self._connection.execute(get_study_query)
      row = study_result.fetchone()
      if not row:
        raise custom_errors.NotFoundError('No such study:', s_resource.name)
      original_study = study_pb2.Study.FromString(row['serialized_study'])

      # Store Study-related metadata into the database.
      datastore.merge_study_metadata(original_study.study_spec, study_metadata)
      update_study_query = (
          sqla.update(self._studies_table)
          .where(self._studies_table.c.study_name == study_name)
          .values(serialized_study=original_study.SerializeToString())
      )
      self._connection.execute(update_study_query)

      # Split the trial-related metadata by Trial.
      split_metadata: DefaultDict[str, List[datastore.UnitMetadataUpdate]] = (
          collections.defaultdict(list)
      )
      for md in trial_metadata:
        split_metadata[md.trial_id].append(md)

      # Now, we update one Trial at a time:
      for trial_id, md_list in split_metadata.items():
        t_resource = s_resource.trial_resource(trial_id)

        # Obtain original trial.
        trial_name = t_resource.name
        original_trial_query = sqla.select([self._trials_table]).where(
            self._trials_table.c.trial_name == trial_name
        )
        trial_result = self._connection.execute(original_trial_query)
        row = trial_result.fetchone()
        if not row:
          raise custom_errors.NotFoundError('No such trial:', trial_name)
        original_trial = study_pb2.Trial.FromString(row['serialized_trial'])

        # Update Trial.
        datastore.merge_trial_metadata(original_trial, md_list)
        update_trial_query = (
            sqla.update(self._trials_table)
            .where(self._trials_table.c.trial_name == trial_name)
            .values(serialized_trial=original_trial.SerializeToString())
        )
        self._connection.execute(update_trial_query)



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Eagle Strategy Optimizer.

Implements a variation of Eagle Strategy without the Levy random walk, aka
Firefly Algorithm (FA) for the purpose of optimizing a given objective function.

Reference: Yang XS. (2009) Firefly Algorithms for Multimodal Optimization.
In: Stochastic Algorithms: Foundations and Applications (SAGA) 2009.
DOI: https://doi.org/10.1007/978-3-642-04944-6_14

Firefly Algorithm Summary
=========================
FA is a genetic algorithm that maintains a pool of fireflies. Each
firefly emits a light whose intensity is non-decreasing in (or simply equal
to) the objective value. Each iteration, a firefly chases after a brighter
firefly, but the brightness it perceives decreases in distance. This allows
multiple "clusters" to form, as opposed to all fireflies collapsing to a
single point. Not included in the original algorithm, we added "repulsion" which
in addition to the "attraction" forces, meaning fireflies move towards the
bright spots as well as away from the dark spots. We also support non-decimal
parameter types (categorical, discrete, integer), and treat them uniquely when
computing distance, adding pertrubation, and mutating fireflies.

For more details, see the linked paper.

OSS Vizier Implementation Summary
=================================
The firefly are stored in three Numpy arrays: features, rewards, perturbations.
Each iteration we mutate 'batch_size' fireflies to generate new features. The
new features are evaluated on the objective function to obtain their associated
rewards and update the pool where improvement was obtained, and decrease the
perturbation factor otherwise.

If the firefly's perturbation reaches the `perturbation_lower_bound` threshold
it's removed and replaced with new random features.

For performance consideration, the 'pool size' is a multiplier of the
'batch size', and so each iteration the pool is sliced to obtain the current
fireflies to be mutated.

Passing prior trials is supported for knowledge-transfering from previous runs
of the optimizer. Prior trials are used to populate the pool with fireflies
which are closer to the optimum.

Note that the strategy assumes that the search space is not conditional.

Example
=======
# Construct the optimizer with eagle strategy.
optimizer = VectorizedOptimizer(
    VectorizedEagleStrategyFactory(),
)
# Run the optimization.
optimizer.optimize(problem_statement, objective_function)

# Access the best features and reward.
best_reward, best_parameters = optimizer.best_results
"""

import logging
from typing import Optional, Literal

import attr
import numpy as np
from vizier._src.algorithms.optimizers import eagle_param_handler
from vizier._src.algorithms.optimizers import vectorized_base as vb
from vizier.pyvizier import converters


@attr.define
class EagleStrategyConfig:
  """Eagle Strategy optimizer config.

  Attributes:
    visibility: The sensetivity to distance between flies when computing pulls.
    gravity: The maximum amount of attraction pull.
    negative_gravity: The maximum amount of repulsion pull.
    perturbation: The default amount of noise for perturbation.
    categorical_perturbation_factor: A factor to apply on categorical params.
    pure_categorical_perturbation_factor: A factor on purely categorical space.
    perturbation_lower_bound: The threshold below flies are removed from pool.
    penalize_factor: The perturbation decrease for unsuccessful flies.
    pool_size_exponent: An exponent for computing pool size based on search
      space size.
    pool_size: An optional way to set the pool size. If not 0, this takes
      precedent over the programatic pool size computation.
    mutate_normalization_type: The type of force mutation normalizatoin used for
      damping down the force intensity to stay within reasonable bounds.
    normalization_scale: A mutation force scale-factor to control intensity.
    prior_trials_pool_pct: The percentage of the pool populated by prior trials.
  """

  # Visibility
  visibility: float = 0.45
  # Gravity
  gravity: float = 1.5
  negative_gravity: float = 0.008
  # Perturbation
  perturbation: float = 0.16
  categorical_perturbation_factor: float = 25
  pure_categorical_perturbation_factor: float = 30
  # Penalty
  perturbation_lower_bound: float = 7e-5
  penalize_factor: float = 7e-1
  # Pool size
  pool_size_exponent: float = 1.2
  pool_size: int = 0
  # Force normalization mode
  mutate_normalization_type: str = "mean"
  # Multiplier factor when using normalized modes.
  normalization_scale: float = 0.5
  # The percentage of the firefly pool to be populated with prior trials.
  prior_trials_pool_pct: float = 0.96


@attr.define(frozen=True)
class VectorizedEagleStrategyFactory(vb.VectorizedStrategyFactory):
  """Eagle strategy factory."""

  eagle_config: EagleStrategyConfig = attr.field(factory=EagleStrategyConfig)

  def __call__(
      self,
      converter: converters.TrialToArrayConverter,
      suggestion_batch_size: int = 5,
      seed: Optional[int] = None,
      prior_features: Optional[np.ndarray] = None,
      prior_rewards: Optional[np.ndarray] = None,
  ) -> "VectorizedEagleStrategy":
    """Create a new vectorized eagle strategy.

    In order to create the strategy a converter has to be passed, which the
    strategy will then store a pointer to. The strategy uses the converter to
    get information about the original Vizier Parameters and their relation to
    the array indices (which index belong to which Vizier Parameter). This is
    useful for example for sampling CATEGORICAL features.

    Arguments:
      converter: TrialToArrayConverter that matches the converter used in
        computing the objective / acuisition function.
      suggestion_batch_size: The batch_size of the returned suggestion array.
      seed: The seed to input into the random generator.
      prior_features: The prior features for populating the pool.
        (n_prior_candidates, n_features)
      prior_rewards:  The prior rewards for populating the pool.
        (n_prior_candidates,)

    Returns:
      A new instance of VectorizedEagleStrategy.
    """
    return VectorizedEagleStrategy(
        converter=converter,
        config=self.eagle_config,
        batch_size=suggestion_batch_size,
        seed=seed,
        prior_features=prior_features,
        prior_rewards=prior_rewards,
    )


# TODO: Create jit-compatible JAX version.
@attr.define
class VectorizedEagleStrategy(vb.VectorizedStrategy):
  """Eagle strategy implementation for maximization problem based on Numpy.

  Attributes:
    converter: The converter used for the optimization problem.
    config: The Eagle strategy configuration.
    n_features: The number of features.
    batch_size: The number of suggestions generated at each suggestion call.
    prior_features: The prior features to be used for seeding the pool. When the
      optimizer is used to optimize a designer's acquisition function, the prior
      features are the previous designer suggestions provided in the ordered
      they were suggested. Shape: (n_prior_trials, n_features).
    prior_rewards: The associated prior rewards of the prior features.
      Shape: (n_prior_trials,)
    seed: The seed to generate random values.
    pool_size: The total number of flies in the pool.
    _features: Array with dimensions (suggestion_count, feature_count) storing
      the firefly features.
    _rewards: Array with dimensions (suggestion_count,) that for each firefly
      stores the current (best) associated objective function result.
    _perturbations: Array with dimensions (suggestion_count,) storing the
      firefly current perturbations.
    _batch_id: The current batch index which is in [0, pool_size/batch_size - 1]
    _batch_slice: The slice of the indices in the current batch.
    _iterations: The total number of batches suggested.
    _num_removed_flies: The total number of removed flies (for debugging).
    _best_results: An Heap storing the best results across all flies.
    _last_suggested_features: The last features the strategy has suggested.
  """

  converter: converters.TrialToArrayConverter
  config: EagleStrategyConfig = attr.field(init=True, repr=False)
  batch_size: int = attr.field(init=True)
  seed: Optional[int] = attr.field(init=True)
  pool_size: int = attr.field(init=False)
  prior_features: Optional[np.ndarray] = attr.field(init=True, default=None)
  prior_rewards: Optional[np.ndarray] = attr.field(init=True, default=None)
  # Attributes related to the strategy's state.
  _features: np.ndarray = attr.field(init=False, repr=False)
  _rewards: np.ndarray = attr.field(init=False, repr=False)
  _perturbations: np.ndarray = attr.field(init=False, repr=False)
  _batch_id: int = attr.field(init=False, repr=False)
  _batch_slice: slice = attr.field(init=False, repr=False)
  _iterations: int = attr.field(init=False)
  _last_suggested_features: np.ndarray = attr.field(init=False, repr=False)
  _best_reward: float = attr.field(init=False)
  # Attributes related to computations.
  _num_removed_flies: int = attr.field(init=False, default=0)
  _n_features: int = attr.field(init=False)
  _perturbation_factors: np.ndarray = attr.field(init=False, repr=False)

  def __attrs_post_init__(self):
    self._initialize()
    logging.info("Eagle class attributes:\n%s", self)
    logging.info("Eagle configuration:\n%s", self.config)

  def _compute_pool_size(self):
    """Compute the pool size, and ensures it's a multiply of the batch_size."""
    raw_pool_size = 10 + int(
        0.5 * self._n_features
        + self._n_features**self.config.pool_size_exponent
    )
    return int(np.ceil(raw_pool_size / self.batch_size) * self.batch_size)

  def _initialize(self):
    """Initialize the designer state."""
    self._rng = np.random.default_rng(self.seed)
    self._param_handler = eagle_param_handler.EagleParamHandler(
        converter=self.converter,
        rng=self._rng,
        categorical_perturbation_factor=self.config.categorical_perturbation_factor,
        pure_categorical_perturbation_factor=self.config.pure_categorical_perturbation_factor,
    )
    self._n_features = self._param_handler.n_features
    if self.config.pool_size > 0:
      # This allow to override the pool size computation.
      self.pool_size = self.config.pool_size
    else:
      self.pool_size = self._compute_pool_size()
    logging.info("Pool size: %d", self.pool_size)
    if self.batch_size == -1:
      # This configuration updates all the fireflies in each iteration.
      self.batch_size = self.pool_size
    self._batch_id = 0
    self._iterations = 0
    self._batch_slice = np.s_[0 : self.batch_size]
    self._perturbations = (
        np.ones(
            self.pool_size,
        )
        * self.config.perturbation
    )
    self._last_suggested_features = None
    self._perturbation_factors = self._param_handler.perturbation_factors
    # Use priors to populate Eagle state
    if self.prior_features is not None or self.prior_rewards is not None:
      self._populate_pool_with_prior_trials()
    else:
      self._features = self._param_handler.random_features(
          self.pool_size, self._n_features
      )
    # Rewards are not populated from seed, as they were potentially generated
    # from a different objective function. New rewards will be obtained.
    self._rewards = np.ones(
        self.pool_size,
    ) * (-np.inf)
    self._best_reward = -np.inf
    # Ignore subtracting np.inf - np.inf. See '_compute_scaled_directions' for
    # explanation on why we ignore warning in this case.
    np.seterr(invalid="ignore")

  def _populate_pool_with_prior_trials(self) -> None:
    """Populate the pool with prior trials.

    A portion of the pool is first populated with random features based on
    'prior_trials_pool_pct', then the rest of the flies are populated by
    sequentially iterate over the prior trials, finding the cloest firefly in
    the pool and replace it if the reward is better.
    """
    if self.prior_features is None or self.prior_rewards is None:
      raise ValueError("One of prior features / prior rewards wasn't provided!")
    if self.prior_features.shape[0] != self.prior_rewards.shape[0]:
      raise ValueError(
          f"prior features shape ({self.prior_features.shape[0]}) doesn't match"
          f" prior  rewards shape ({self.prior_rewards.shape[0]})!"
      )
    if self.prior_features.shape[1] != self._n_features:
      raise ValueError(
          "prior features shape doesn't match n_features{self._n_features}!"
      )
    if len(self.prior_rewards.shape) > 1:
      raise ValueError("prior rewards is expected to be 1D array!")

    # Reverse the order of prior trials to assign more weight to recent trials.
    self.prior_features = np.flip(self.prior_features, axis=-1)
    self.prior_rewards = np.flip(self.prior_rewards, axis=-1)

    self._features = np.zeros((0, self._n_features))
    # Fill pool with random features.
    n_random_flies = int(
        self.pool_size * (1 - self.config.prior_trials_pool_pct)
    )
    self._features = self._param_handler.random_features(
        n_random_flies, self._n_features
    )
    pool_left_space = self.pool_size - n_random_flies

    if self.prior_features.shape[0] < pool_left_space:
      # Less prior trials than left space. Take all prior trials for the pool.
      self._features = np.concatenate([self._features, self.prior_features])
      # Randomize the rest of the pool fireflies.
      random_features = self._param_handler.random_features(
          self.pool_size - len(self._features), self._n_features
      )
      self._features = np.concatenate([self._features, random_features])
    else:
      # More prior trials than left space. Iteratively populate the pool.
      tmp_features = self.prior_features[:pool_left_space]
      tmp_rewards = self.prior_rewards[:pool_left_space]
      for i in range(pool_left_space, self.prior_features.shape[0]):
        ind = np.argmin(
            np.sum(np.square(self.prior_features[i] - tmp_features), axis=-1)
        )
        if tmp_rewards[ind] < self.prior_rewards[i]:
          # Only take the prior trials features. Rewards obtain during update.
          tmp_features[ind] = self.prior_features[i]
          tmp_rewards[ind] = self.prior_rewards[i]
      self._features = np.concatenate([self._features, tmp_features])

  @property
  def suggestion_batch_size(self) -> int:
    """The number of suggestions returned at each call of 'suggest'."""
    return self.batch_size

  def suggest(self) -> np.ndarray:
    """Suggest new mutated and perturbed features.

    After initializing, at each call `batch_size` fireflies are mutated to
    generate new features using pulls (attraction/repulsion) from all other
    fireflies in the pool.

    Returns:
      suggested batch features: (batch_size, n_features)
    """
    if self._iterations < self.pool_size // self.batch_size:
      # The strategy is still initializing. Return the random/prior features.
      new_features = self._features[self._batch_slice]
    else:
      mutated_features = self._create_features()
      perturbations = self._create_perturbations()
      new_features = mutated_features + perturbations

    new_features = self._param_handler.sample_categorical(new_features)
    suggested_features = np.clip(new_features, 0, 1)
    # Save the suggested features to be used in update.
    self._last_suggested_features = suggested_features
    return suggested_features

  def _increment_batch(self):
    """Increment the batch of fireflies features are generate from."""
    self._batch_id = (self._batch_id + 1) % (self.pool_size // self.batch_size)
    start_batch = self._batch_id * self.batch_size
    end_batch = (self._batch_id + 1) * self.batch_size
    self._batch_slice = np.s_[start_batch:end_batch]

  def _create_features(self) -> np.ndarray:
    """Create new batch of mutated and perturbed features.

    Returns:
      batch features: (batch_size, n_features)
    """
    features_diffs, dists = self._compute_features_diffs_and_dists()
    scaled_directions = self._compute_scaled_directions()
    features_changes = self._compute_features_changes(
        features_diffs, dists, scaled_directions
    )
    return self._features[self._batch_slice] + features_changes

  def _compute_features_diffs_and_dists(self) -> tuple[np.ndarray, np.ndarray]:
    """Compute the features difference and distances.

    The computation is done between the 'batch_size' fireflies and all
    other fireflies in the pool.

    features_diff[i, j, :] := features[j, :] - features[i, :]
    features_dist[i, j, :] := distance between fly 'j' and fly 'i'

    Returns:
      feature differences: (batch_size, pool_size, n_features)
      features distances: (batch_size, pool_size)
    """
    shape = (self.batch_size,) + self._features.shape
    features_diffs = np.broadcast_to(self._features, shape) - np.expand_dims(
        self._features[self._batch_slice], 1
    )
    dists = np.sum(np.square(features_diffs), axis=-1)
    return features_diffs, dists

  def _compute_scaled_directions(self) -> np.ndarray:
    """Compute the scaled direction for applying pull between two flies.

    scaled_directions[i,j] := direction of force applied by fly 'j' on fly 'i'.

    Note that to compute 'directions' we might perform subtract with removed
    flies with having value of -np.inf. Moreover, we might even subtract between
    two removed flies which will result in np.nan. Both cases are handled when
    computing the actual feautre changes applying a relevant mask.

    Returns:
      scaled directions: (batch_size, pool_size)
    """
    shape = (self.batch_size,) + self._rewards.shape
    directions = np.broadcast_to(self._rewards, shape) - np.expand_dims(
        self._rewards[self._batch_slice], -1
    )

    scaled_directions = np.where(
        directions >= 0, self.config.gravity, -self.config.negative_gravity
    )
    return scaled_directions

  def _compute_features_changes(
      self,
      features_diffs: np.ndarray,
      dists: np.ndarray,
      scaled_directions: np.ndarray,
  ) -> np.ndarray:
    """Compute the firefly features changes due to mutation.

    The pool fireflies forces (pull/push) are being normalized to ensure the
    combined force doesn't throw the firefly too far. Mathematically, the
    normalization guarantees that the combined normalized force is within the
    simplex constructed by the unnormalized forces and therefore within bounds.

    Arguments:
      features_diffs: (batch_size, pool_size, n_features)
      dists: (batch_size, pool_size)
      scaled_directions: (batch_size, pool_size)

    Returns:
      feature changes: (batch_size, feature_n)
    """
    # Normalize the distance by the number of features.
    force = np.exp(-self.config.visibility * dists / self._n_features * 10)
    scaled_force = np.expand_dims(scaled_directions * force, -1)
    # Handle removed fireflies without updated rewards.
    inf_indx = np.isinf(self._rewards)
    if np.sum(inf_indx) == self.pool_size:
      logging.warning(
          (
              "All firefly were recently removed. This Shouldn't happen."
              "Pool Features:\n%sPool rewards:\n%s"
          ),
          self._features,
          self._rewards,
      )
      return np.zeros((self.batch_size, self._n_features))

    # Ignore fireflies that were removed from the pool.
    features_diffs = features_diffs[:, ~inf_indx, :]
    scaled_force = scaled_force[:, ~inf_indx, :]

    # Separate forces to pull and push so to normalize them separately.
    scaled_pulls = np.where(scaled_force > 0, scaled_force, 0)
    scaled_push = np.where(scaled_force < 0, scaled_force, 0)

    if self.config.mutate_normalization_type == "mean":
      # Divide the push and pull forces by the number of flies participating.
      # Also multiply by normalization_scale.
      norm_scaled_pulls = self.config.normalization_scale * np.nan_to_num(
          scaled_pulls / np.sum(scaled_pulls > 0, axis=1, keepdims=True), 0
      )
      norm_scaled_push = self.config.normalization_scale * np.nan_to_num(
          scaled_push / np.sum(scaled_push < 0, axis=1, keepdims=True), 0
      )
    elif self.config.mutate_normalization_type == "random":
      # Create random matrices and normalize each row, s.t. the sum is 1.
      pull_rand_matrix = self._rng.uniform(
          0, 1, size=scaled_pulls.shape
      ) * np.int_(scaled_pulls > 0)
      pull_weight_matrix = pull_rand_matrix / np.sum(
          pull_rand_matrix, axis=1, keepdims=True
      )
      push_rand_matrix = self._rng.uniform(
          0, 1, size=scaled_pulls.shape
      ) * np.int_(scaled_pulls > 0)
      push_weight_matrix = push_rand_matrix / np.sum(
          push_rand_matrix, axis=1, keepdims=True
      )
      # Normalize pulls/pulls by the weight matrices.
      # Also multiply by normalization_scale.
      norm_scaled_pulls = (
          self.config.normalization_scale * scaled_pulls * pull_weight_matrix
      )
      norm_scaled_push = (
          self.config.normalization_scale * scaled_push * push_weight_matrix
      )
    elif self.config.mutate_normalization_type == "unnormalized":
      # Doesn't normalize the forces. Use this option with caution.
      norm_scaled_pulls = scaled_pulls
      norm_scaled_push = scaled_push

    # Sums normalized forces (pull/push) of all fireflies.
    return np.sum(
        features_diffs * (norm_scaled_pulls + norm_scaled_push), axis=1
    )

  def _create_perturbations(self) -> np.ndarray:
    """Create random perturbations for the newly creatd batch.

    Returns:
      batched perturbations: (base_size, n_features)
    """
    # Generate normalized noise for each batch.
    batch_noise = self._rng.laplace(size=(self.batch_size, self._n_features))
    batch_noise /= np.max(np.abs(batch_noise), axis=1, keepdims=True)
    # Scale the noise by the each fly current perturbation.
    return (
        batch_noise
        * self._perturbations[self._batch_slice][:, np.newaxis]
        * self._perturbation_factors
    )

  def update(self, batch_rewards: np.ndarray) -> None:
    """Update the firefly pool based on the new batch of results.

    Arguments:
      batch_rewards: (batch_size, )
    """
    self._update_best_reward(batch_rewards)
    if self._iterations < self.pool_size // self.batch_size:
      # The strategy is still initializing. Assign rewards.
      self._features[self._batch_slice] = self._last_suggested_features
      self._rewards[self._batch_slice] = batch_rewards
    else:
      # Pass the new batch rewards and the associated last suggested features.
      self._update_pool_features_and_rewards(batch_rewards)
      self._trim_pool()
    self._increment_batch()
    self._iterations += 1

  def _update_best_reward(self, batch_rewards: np.ndarray) -> None:
    """Store the best result seen thus far to be used in pool trimming."""
    self._best_reward = np.max([self._best_reward, np.max(batch_rewards)])

  def _update_pool_features_and_rewards(
      self,
      batch_rewards: np.ndarray,
  ):
    """Update the features and rewards for flies with improved rewards.

    Arguments:
      batch_rewards: (batch_size, )
    """
    sliced_features = self._features[self._batch_slice]
    sliced_rewards = self._rewards[self._batch_slice]
    sliced_perturbations = self._perturbations[self._batch_slice]
    # Find indices of flies that their generated features made an improvement.
    improve_indx = batch_rewards > sliced_rewards
    # Update successful flies' with the associated last features and rewards.
    sliced_features[improve_indx] = self._last_suggested_features[improve_indx]
    sliced_rewards[improve_indx] = batch_rewards[improve_indx]
    # Penalize unsuccessful flies.
    sliced_perturbations[~improve_indx] *= self.config.penalize_factor

  def _trim_pool(self) -> None:
    """Trim the pool by replacing unsuccessful fireflies with new random ones.

    A firefly is considered unsuccessful if its current perturbation is below
    'perturbation_lower_bound' and it's not the best fly seen thus far.
    Random features are created to replace the existing ones, and rewards
    are set to -np.inf to indicate that we don't have values for those feaures
    yet and we shouldn't use them during suggest.
    """
    sliced_perturbations = self._perturbations[self._batch_slice]
    indx = sliced_perturbations < self.config.perturbation_lower_bound
    n_remove = np.sum(indx)
    if n_remove > 0:
      sliced_features = self._features[self._batch_slice]
      sliced_rewards = self._rewards[self._batch_slice]
      # Ensure the best firefly is never removed. For optimization purposes,
      # this logic is inside the if statement to be peformed only if needed.
      indx = indx & (sliced_rewards != self._best_reward)
      n_remove = np.sum(indx)
      if n_remove == 0:
        return
      # Replace fireflies with random features and evaluate rewards.
      sliced_features[indx] = self._param_handler.random_features(
          n_remove, self._n_features
      )
      sliced_perturbations[indx] = (
          np.ones(
              n_remove,
          )
          * self.config.perturbation
      )
      # Setting rewards to -inf to filter out those fireflies during suggest.
      sliced_rewards[indx] = np.ones(
          n_remove,
      ) * (-np.inf)
      self._num_removed_flies += n_remove



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for vectorized_base."""

import datetime
import time
from typing import Optional

import mock
import numpy as np
from vizier import pyvizier as vz
from vizier._src.algorithms.optimizers import vectorized_base as vb
from vizier.pyvizier import converters

from absl.testing import absltest
from absl.testing import parameterized


class FakeIncrementVectorizedStrategy(vb.VectorizedStrategy):
  """Fake vectorized strategy with incrementing suggestions."""

  def __init__(self, *args, **kwargs):
    self._iter = 0

  def suggest(self) -> np.ndarray:
    # The following structure allows to test the top K results.
    suggestions = (
        np.array([
            [self._iter % 10, self._iter % 10],
            [(self._iter + 1) % 10, (self._iter + 1) % 10],
            [(self._iter + 2) % 10, (self._iter + 2) % 10],
            [(self._iter + 3) % 10, (self._iter + 3) % 10],
            [(self._iter + 4) % 10, (self._iter + 4) % 10],
        ])
        / 10
    )
    self._iter += 5
    return suggestions

  @property
  def suggestion_batch_size(self) -> int:
    return 5

  def update(self, rewards: np.ndarray) -> None:
    pass


# pylint: disable=unused-argument
def fake_increment_strategy_factory(
    converter: converters.TrialToArrayConverter,
    suggestion_batch_size: int,
    seed: Optional[int] = None,
    prior_features: Optional[np.ndarray] = None,
    prior_rewards: Optional[np.ndarray] = None,
) -> vb.VectorizedStrategy:
  return FakeIncrementVectorizedStrategy()


class FakePriorTrialsVectorizedStrategy(vb.VectorizedStrategy):
  """Fake vectorized strategy to test prior trials."""

  def __init__(self, prior_features: np.ndarray, prior_rewards: np.ndarray):
    self.seed_features, self.seed_rewards = prior_features, prior_rewards
    if len(self.seed_rewards.shape) != 1:
      raise ValueError('Expected seed labels to have 1D dimension!')

  def suggest(self) -> np.ndarray:
    return self.seed_features[np.argmax(self.seed_rewards, axis=-1)].reshape(
        1, -1
    )

  @property
  def suggestion_batch_size(self) -> int:
    return 1

  def update(self, rewards: np.ndarray) -> None:
    pass


# pylint: disable=unused-argument
def fake_prior_trials_strategy_factory(
    converter: converters.TrialToArrayConverter,
    suggestion_batch_size: int,
    seed: Optional[int] = None,
    prior_features: Optional[np.ndarray] = None,
    prior_rewards: Optional[np.ndarray] = None,
) -> vb.VectorizedStrategy:
  return FakePriorTrialsVectorizedStrategy(prior_features, prior_rewards)


class VectorizedBaseTest(parameterized.TestCase):

  @parameterized.parameters(1, 2, 3, 4, 5, 6, 7, 8, 9, 10)
  def test_optimize_candidates_len(self, count):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('f1', 0.0, 10.0)
    problem.search_space.root.add_float_param('f2', 0.0, 10.0)
    converter = converters.TrialToArrayConverter.from_study_config(problem)
    score_fn = lambda x: np.sum(x, axis=-1)
    optimizer = vb.VectorizedOptimizer(
        strategy_factory=fake_increment_strategy_factory,
        max_evaluations=100,
    )
    res = optimizer.optimize(
        converter=converter, score_fn=score_fn, count=count
    )
    self.assertLen(res, count)

  @parameterized.parameters([None, datetime.timedelta(minutes=10)])
  def test_should_stop_max_evaluations(self, max_duration):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('f1', 0.0, 10.0)
    problem.search_space.root.add_float_param('f2', 0.0, 10.0)
    converter = converters.TrialToArrayConverter.from_study_config(problem)
    score_fn = mock.Mock()
    score_fn.side_effect = lambda x: np.sum(x, axis=-1)
    optimizer = vb.VectorizedOptimizer(
        strategy_factory=fake_increment_strategy_factory,
        max_evaluations=100,
        max_duration=max_duration,
    )
    optimizer.optimize(converter=converter, score_fn=score_fn, count=3)
    # The batch size is 5, so we expect 100/5 = 20 calls
    self.assertEqual(score_fn.call_count, 20)

  def test_best_candidates_count_is_1(self):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('f1', 0.0, 1.0)
    problem.search_space.root.add_float_param('f2', 0.0, 1.0)
    converter = converters.TrialToArrayConverter.from_study_config(problem)
    score_fn = lambda x: -np.max(np.square(x - 0.52), axis=-1)
    strategy_factory = FakeIncrementVectorizedStrategy
    optimizer = vb.VectorizedOptimizer(
        strategy_factory=strategy_factory, max_evaluations=10
    )
    best_candidates = optimizer.optimize(
        converter=converter, score_fn=score_fn, count=1
    )
    # check the best candidate
    self.assertEqual(best_candidates[0].parameters['f1'].value, 0.5)
    self.assertEqual(best_candidates[0].parameters['f2'].value, 0.5)
    self.assertEqual(
        best_candidates[0].final_measurement.metrics['acquisition'].value,
        -((0.5 - 0.52) ** 2),
    )

  def test_best_candidates_count_is_3(self):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('f1', 0.0, 1.0)
    problem.search_space.root.add_float_param('f2', 0.0, 1.0)
    converter = converters.TrialToArrayConverter.from_study_config(problem)
    score_fn = lambda x: -np.max(np.square(x - 0.52), axis=-1)
    optimizer = vb.VectorizedOptimizer(
        strategy_factory=fake_increment_strategy_factory, max_evaluations=10
    )
    best_candidates = optimizer.optimize(
        converter=converter, score_fn=score_fn, count=3
    )
    # check 1st best candidate
    self.assertEqual(best_candidates[0].parameters['f1'].value, 0.5)
    self.assertEqual(best_candidates[0].parameters['f2'].value, 0.5)
    self.assertEqual(
        best_candidates[0].final_measurement.metrics['acquisition'].value,
        -((0.5 - 0.52) ** 2),
    )
    # check 2nd best candidate
    self.assertEqual(best_candidates[1].parameters['f1'].value, 0.6)
    self.assertEqual(best_candidates[1].parameters['f2'].value, 0.6)
    self.assertEqual(
        best_candidates[1].final_measurement.metrics['acquisition'].value,
        -((0.6 - 0.52) ** 2),
    )
    # check 3rd best candidate
    self.assertEqual(best_candidates[2].parameters['f1'].value, 0.4)
    self.assertEqual(best_candidates[2].parameters['f2'].value, 0.4)
    self.assertEqual(
        best_candidates[2].final_measurement.metrics['acquisition'].value,
        -((0.4 - 0.52) ** 2),
    )

  def test_should_stop_max_duration(self):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('f1', 0.0, 10.0)
    problem.search_space.root.add_float_param('f2', 0.0, 10.0)
    converter = converters.TrialToArrayConverter.from_study_config(problem)
    score_fn = mock.Mock()

    def slow_score_fn(x):
      time.sleep(1)
      return np.sum(x, axis=-1)

    score_fn.side_effect = slow_score_fn
    optimizer = vb.VectorizedOptimizer(
        strategy_factory=fake_increment_strategy_factory,
        max_evaluations=100,
        max_duration=datetime.timedelta(seconds=3),
    )
    optimizer.optimize(converter=converter, score_fn=score_fn)
    # Test the optimization stopped after ~3 seconds based on function calls.
    self.assertLess(score_fn.call_count, 4)

  def test_vectorized_optimizer_factory(self):
    optimizer_factory = vb.VectorizedOptimizerFactory(
        strategy_factory=fake_increment_strategy_factory
    )
    optimizer = optimizer_factory(suggestion_batch_size=5, max_evaluations=1000)
    self.assertEqual(optimizer.max_evaluations, 1000)
    self.assertEqual(optimizer.suggestion_batch_size, 5)

  def test_prior_trials(self):
    """Test that the optimizer can correctly parsae and pass seed trials."""
    optimizer_factory = vb.VectorizedOptimizerFactory(
        strategy_factory=fake_prior_trials_strategy_factory
    )
    optimizer = optimizer_factory(suggestion_batch_size=5, max_evaluations=100)

    study_config = vz.ProblemStatement(
        metric_information=[
            vz.MetricInformation(
                name='obj', goal=vz.ObjectiveMetricGoal.MAXIMIZE
            )
        ]
    )
    root = study_config.search_space.root
    root.add_float_param('x1', 0.0, 10.0)
    root.add_float_param('x2', 0.0, 10.0)
    converter = converters.TrialToArrayConverter.from_study_config(study_config)

    trial1 = vz.Trial(parameters={'x1': 1, 'x2': 1})
    measurement1 = vz.Measurement(metrics={'obj': vz.Metric(value=-10.33)})
    trial1.complete(measurement1, inplace=True)

    trial2 = vz.Trial(parameters={'x1': 2, 'x2': 2})
    measurement2 = vz.Measurement(metrics={'obj': vz.Metric(value=5.0)})
    trial2.complete(measurement2, inplace=True)

    best_trial = optimizer.optimize(
        converter,
        lambda x: -np.max(np.square(x - 0.52), axis=-1),
        count=1,
        prior_trials=[trial1, trial2, trial1],
    )
    self.assertEqual(best_trial[0].parameters['x1'].value, 2)
    self.assertEqual(best_trial[0].parameters['x2'].value, 2)

    best_trial = optimizer.optimize(
        converter,
        lambda x: -np.max(np.square(x - 0.52), axis=-1),
        count=1,
        prior_trials=[trial1],
    )
    self.assertEqual(best_trial[0].parameters['x1'].value, 1)
    self.assertEqual(best_trial[0].parameters['x2'].value, 1)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for stochastic_process_model."""

import functools

from absl.testing import parameterized
from flax import linen as nn

import jax
from jax import numpy as jnp
from jax import random
from jax.config import config
import numpy as np
from tensorflow_probability.substrates import jax as tfp
import tree
from vizier._src.jax import stochastic_process_model as sp_model
from absl.testing import absltest

config.update('jax_enable_x64', True)

tfb = tfp.bijectors
tfd = tfp.distributions
tfpk = tfp.math.psd_kernels


def _test_coroutine(inputs=None, dtype=np.float64):
  """A coroutine that follows the `ModelCoroutine` protocol."""
  constraint = sp_model.Constraint(
      bounds=(np.zeros([], dtype=dtype), None), bijector=tfb.Exp()
  )
  amplitude = yield sp_model.ModelParameter(
      init_fn=lambda k: random.exponential(k, dtype=dtype),
      regularizer=lambda x: dtype(1e-3) * x**2,
      constraint=constraint,
      name='amplitude',
  )
  inverse_length_scale = yield sp_model.ModelParameter.from_prior(
      tfd.Exponential(
          rate=np.ones([], dtype=dtype), name='inverse_length_scale'
      ),
      constraint=constraint,
  )
  kernel = tfpk.ExponentiatedQuadratic(
      amplitude=amplitude,
      inverse_length_scale=inverse_length_scale,
      validate_args=True)
  return tfd.StudentTProcess(
      df=dtype(5.0), kernel=kernel, index_points=inputs, validate_args=True
  )


def _make_inputs(key, dtype):
  obs_key, pred_key = random.split(key)
  dim = 3
  num_observed = 20
  x_observed = random.uniform(obs_key, shape=(num_observed, dim), dtype=dtype)
  y_observed = x_observed.sum(axis=-1)
  x_predictive = random.uniform(pred_key, shape=(100, 5, dim), dtype=dtype)
  return x_observed, y_observed, x_predictive


class StochasticProcessModelTest(parameterized.TestCase):

  @parameterized.named_parameters(
      # TODO: Add a test case with categorical data.
      # TODO: Fix support for f32.
      {
          'testcase_name': 'continuous_only',
          'model_coroutine': _test_coroutine,
          'test_data_fn': _make_inputs,
          'dtype': np.float64,
      },
  )
  def test_stochastic_process_model(self, model_coroutine, test_data_fn, dtype):
    init_key, data_key, sample_key = jax.random.split(random.PRNGKey(0), num=3)
    x_observed, y_observed, x_predictive = test_data_fn(data_key, dtype)
    model = sp_model.StochasticProcessModel(
        coroutine=functools.partial(model_coroutine, dtype=dtype)
    )

    init_state = model.init(init_key, x_observed)
    dist, losses = model.apply(init_state, x_observed, mutable=('losses',))
    lp = dist.log_prob(y_observed)
    self.assertTrue(np.isfinite(lp))
    self.assertEqual(lp.dtype, dtype)

    # Check regularization loss values.
    gen = model_coroutine(dtype=dtype)
    p = next(gen)
    params = dict(init_state['params'])
    losses = dict(losses['losses'])
    try:
      while True:
        value = params[p.name]
        param_loss = losses[f'{p.name}_regularization']
        self.assertTrue(np.isfinite(value))
        self.assertEqual(value.dtype, dtype)
        self.assertTrue(np.isfinite(param_loss))
        self.assertAlmostEqual(p.regularizer(value), param_loss)
        p = gen.send(value)
    except StopIteration:
      pass

    _, pp_state = model.apply({'params': params},
                              x_observed,
                              y_observed,
                              method=model.precompute_predictive,
                              mutable=('predictive',))

    state = {'params': params, **pp_state}
    pp_dist = model.apply(state, x_predictive, method=model.predict)
    pp_dist2 = model.apply(
        state,
        jax.tree_util.tree_map(lambda x: x + jnp.ones_like(x), x_predictive),
        method=model.predict,
    )

    pp_log_prob = pp_dist.log_prob(pp_dist.sample(seed=sample_key))
    self.assertTrue(np.isfinite(pp_log_prob).all())
    self.assertEqual(pp_log_prob.dtype, dtype)

    # Assert no Cholesky recomputation.
    def _get_cholesky(kernel):
      if isinstance(kernel, tfpk.SchurComplement):
        return kernel._precomputed_divisor_matrix_cholesky
      return kernel.schur_complement._precomputed_divisor_matrix_cholesky

    self.assertIs(_get_cholesky(pp_dist.kernel), _get_cholesky(pp_dist2.kernel))

    # pylint: disable=g-long-lambda
    posterior_fn = jax.jit(
        lambda x: model.apply(
            {'params': params},
            x,
            x_observed,
            y_observed,
            mutable=('predictive',),
            method=model.posterior,
        )[0]
    )
    pp_dist3 = posterior_fn(x_predictive)
    pp_dist4 = posterior_fn(x_predictive)
    # Assert no Cholesky recomputation.
    pp_dist3_cholesky = _get_cholesky(pp_dist3.kernel)
    pp_dist4_cholesky = _get_cholesky(pp_dist4.kernel)
    if pp_dist3_cholesky is None and pp_dist4_cholesky is None:
      self.assertIs(pp_dist3_cholesky, pp_dist4_cholesky)
    else:
      self.assertEqual(pp_dist3_cholesky.unsafe_buffer_pointer(),
                       pp_dist4_cholesky.unsafe_buffer_pointer())

  def test_stochastic_process_model_with_mean_fn(self):

    obs_key, pred_key, init_key, vmap_key = random.split(
        random.PRNGKey(0), num=4)
    dim = 3
    num_observed = 20
    x_observed = random.uniform(obs_key, shape=(num_observed, dim))
    y_observed = x_observed.sum(axis=-1)
    x_predictive = random.uniform(pred_key, shape=(100, 5, dim))

    # Use `jnp.squeeze` to remove the singleton dimension of the output of
    # `mean_fn`, so that it has shape `(num_observed,)`.
    mean_fn = nn.Sequential([nn.Dense(5), nn.Dense(1), jnp.squeeze])
    model = sp_model.StochasticProcessModel(
        coroutine=_test_coroutine, mean_fn=mean_fn)

    init_state = model.init(init_key, x_observed)
    stp = model.apply(init_state, x_observed, mutable=False)
    lp = stp.log_prob(y_observed)
    self.assertTrue(np.isfinite(lp))

    # The mean of the GP should equal `mean_fn` evaluated at the index points.
    np.testing.assert_allclose(
        stp.mean(),
        mean_fn.apply({'params': init_state['params']['mean_fn']}, x_observed))

    _, pp_state = model.apply({'params': init_state['params']},
                              x_observed,
                              y_observed,
                              method=model.precompute_predictive,
                              mutable=('predictive',))

    state = {'params': init_state['params'], **pp_state}
    pp_dist = model.apply(state, x_predictive, method=model.predict)
    pp_mean = pp_dist.mean()

    self.assertTrue(
        np.isfinite(pp_dist.log_prob(x_predictive.sum(axis=-1))).all())
    self.assertSequenceEqual(pp_mean.shape, (100, 5))
    self.assertTrue(np.isfinite(pp_mean).all())

    # Test that vmap works.
    def log_prob(p):
      return model.apply(
          p, x_observed, mutable=('losses',))[0].log_prob(y_observed)

    batch_size = 20
    keys = random.split(vmap_key, num=batch_size)
    params = jax.vmap(lambda k: model.init(k, x_observed))(keys)
    lp = jax.vmap(log_prob)(params)
    self.assertLen(lp, batch_size)
    self.assertTrue(np.isfinite(lp).all())


class VectorToArrayTreeTest(absltest.TestCase):

  def test_vector_to_array_tree(self):
    k0, k1 = random.split(random.PRNGKey(0))
    params = {
        'foo': random.uniform(k0, shape=(2, 3)),
        'bar': np.array(0.3),
        'baz': random.uniform(k1, shape=(6,))
    }
    bijector = sp_model.VectorToArrayTree(params)
    self.assertEqual(bijector.inverse(params).shape, (13,))

    v = np.ones([13])
    struct = bijector.forward(v)
    self.assertSameElements(list(struct.keys()), list(params.keys()))
    for k in params:
      self.assertEqual(struct[k].shape, params[k].shape)


class ModelParameterTest(absltest.TestCase):

  def test_parameter_from_prior(self):
    prior = tfd.Gamma(concentration=1., rate=1.)
    param = sp_model.ModelParameter.from_prior(prior=prior)
    samples = param.init_fn(random.PRNGKey(0))
    regularization = param.regularizer(samples)
    x = np.linspace(-10.0, 10.0, 20)
    self.assertTrue(np.isfinite(regularization))
    self.assertEmpty(regularization.shape)
    self.assertNotEqual(regularization, 0.)
    self.assertTrue((param.constraint.bijector(x) > 0.0).all())


class ConstraintTest(parameterized.TestCase):

  @parameterized.parameters(
      (-1.0, None),
      (0.0, 10.0),
      (None, 1.0),
      ({'a': -1.0, 'b': 0.0}, {'a': 1.0, 'b': 10.0}),
      ({'a': None, 'b': 0.0}, {'a': 1.0, 'b': None}),
      (None, {'a': 1.0, 'b': 2.0}),
  )
  def test_create_constraint(self, lower, upper):
    make_array = lambda x: None if x is None else jnp.array(x)
    lower = jax.tree_util.tree_map(make_array, lower)
    upper = jax.tree_util.tree_map(make_array, upper)
    constraint = sp_model.Constraint.create(
        bounds=(lower, upper), bijector_fn=tfb.SoftClip
    )
    x_part = jnp.linspace(-5.0, 5.0, 10)
    x = tree.map_structure(lambda _: x_part, lower or upper)
    y = constraint.bijector(x)
    for y_, b_ in zip(tree.flatten(y), tree.flatten(lower)):
      if b_ is not None:
        self.assertTrue((y_ > b_).all())
    for y_, b_ in zip(tree.flatten(y), tree.flatten(upper)):
      if b_ is not None:
        self.assertTrue((y_ < b_).all())

  def test_get_constraints(self):
    constraint = sp_model.get_constraints(_test_coroutine)
    x_part = np.linspace(-5.0, 5.0, 10)
    x = {'amplitude': x_part, 'inverse_length_scale': x_part}
    y = constraint.bijector(x)
    lower, upper = constraint.bounds

    self.assertSameElements(list(x.keys()), list(y.keys()))
    for y_, b_ in zip(tree.flatten(y), tree.flatten(lower)):
      if b_ is not None:
        self.assertTrue((y_ > b_).all())
    for y_, b_ in zip(tree.flatten(y), tree.flatten(upper)):
      if b_ is not None:
        self.assertTrue((y_ < b_).all())


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Collection of well-tuned GP models."""

# TODO: Add Ax/BoTorch GP.

import functools
from typing import Any, Generator

import attr
import chex
import jax
from jax import numpy as jnp
from jax.config import config
import numpy as np
from tensorflow_probability.substrates import jax as tfp
from vizier._src.jax import stochastic_process_model as sp
from vizier._src.jax.optimizers import optimizers

# Jax disables float64 computations by default and will silently convert
# float64s to float32s. We must explicitly enable float64.
config.update('jax_enable_x64', True)

tfb = tfp.bijectors
tfd = tfp.distributions
Array = Any
tfpk = tfp.math.psd_kernels


@attr.define
class VizierGaussianProcess(sp.ModelCoroutine[chex.Array, tfd.GaussianProcess]):
  """Vizier's tuned GP.

  See __call__ method documentation.

  Attributes:
    _boundary_epsilon: We expand the constraints by this number so that the
      values exactly at the boundary can be mapped to unconstrained space. i.e.
      we are trying to avoid SoftClip(low=1e-2, high=1.).inverse(1e-2) giving
      NaN.
  """

  _feature_dim: int
  _use_retrying_cholesky: bool = attr.field(default=True, kw_only=True)
  _boundary_epsilon: float = attr.field(default=1e-12, kw_only=True)

  @classmethod
  def model_and_loss_fn(
      cls,
      features: chex.Array,
      labels: chex.Array,
      *,
      use_retrying_cholesky: bool = True,
  ) -> tuple[sp.StochasticProcessModel, optimizers.LossFunction]:
    """Returns the model and loss function."""
    gp_coroutine = VizierGaussianProcess(
        features.shape[-1], use_retrying_cholesky=use_retrying_cholesky
    )
    model = sp.StochasticProcessModel(gp_coroutine)

    # Run ARD.
    def loss_fn(params):
      gp, mutables = model.apply({'params': params},
                                 features,
                                 mutable=['losses', 'predictive'])
      # Normalize so we can use the same learning rate regardless of
      # how many examples we have.
      loss = (-gp.log_prob(labels) + jax.tree_util.tree_reduce(
          jax.numpy.add, mutables['losses'])) / features.shape[0]
      return loss, dict()

    return model, loss_fn

  def _log_uniform_init(
      self, low: float, high: float, shape: tuple[int,
                                                  ...] = tuple()) -> sp.InitFn:
    r"""Take log-uniform sample in the constraint and map it back to \R.

    Args:
      low: Parameter lower bound.
      high: Parameter upper bound.
      shape: Returned array has this shape. Each entry in the returned array is
        an i.i.d sample.

    Returns:
      Randomly sampled array.
    """

    def sample(key: Any) -> jnp.ndarray:
      unif = jnp.array(jax.random.uniform(key, shape, dtype=jnp.float64))
      return jnp.exp(unif * jnp.log(high / low) + jnp.log(low))

    return sample

  def __call__(
      self,
      inputs: Array = None
  ) -> Generator[sp.ModelParameter, Array, tfd.GaussianProcess]:
    """Creates a generator.

    Args:
      inputs: Floating array of dimension (num_examples, _feature_dim).

    Yields:
      GaussianProcess whose event shape is `num_examples`.
    """

    eps = self._boundary_epsilon
    observation_noise_bounds = (np.float64(1e-10 - eps), 1.0 + eps)
    amplitude_bounds = (np.float64(1e-3 - eps), 10.0 + eps)
    ones = np.ones((self._feature_dim,), dtype=np.float64)
    length_scale_bounds = (ones * (1e-2 - eps), ones * 1e2 + eps)

    signal_variance = yield sp.ModelParameter(
        init_fn=self._log_uniform_init(*amplitude_bounds),
        constraint=sp.Constraint(
            amplitude_bounds,
            tfb.SoftClip(*amplitude_bounds, hinge_softness=1e-2),
        ),
        regularizer=lambda x: 0.01 * jnp.log(x / 0.039) ** 2,
        name='signal_variance',
    )
    kernel = tfpk.MaternFiveHalves(amplitude=jnp.sqrt(signal_variance))

    length_scale = yield sp.ModelParameter(
        init_fn=self._log_uniform_init(
            *length_scale_bounds, shape=(self._feature_dim,)
        ),
        constraint=sp.Constraint(
            length_scale_bounds,
            tfb.SoftClip(*length_scale_bounds, hinge_softness=1e-2),
        ),
        regularizer=lambda x: jnp.sum(0.01 * jnp.log(x / 0.5) ** 2),
        name='length_scale_squared',
    )
    kernel = tfpk.FeatureScaled(kernel, scale_diag=jnp.sqrt(length_scale))

    observation_noise_variance = yield sp.ModelParameter(
        init_fn=self._log_uniform_init(*observation_noise_bounds),
        constraint=sp.Constraint(
            observation_noise_bounds,
            tfb.SoftClip(*observation_noise_bounds, hinge_softness=1e-2),
        ),
        regularizer=lambda x: 0.01 * jnp.log(x / 0.0039) ** 2,
        name='observation_noise_variance',
    )

    cholesky_fn = None
    # When cholesky fails, increase jitters and retry.
    if self._use_retrying_cholesky:
      retrying_cholesky = functools.partial(
          tfp.experimental.distributions.marginal_fns.retrying_cholesky,
          jitter=np.float64(1e-4),
          max_iters=5,
      )
      cholesky_fn = lambda matrix: retrying_cholesky(matrix)[0]

    return tfd.GaussianProcess(
        kernel,
        index_points=inputs,
        observation_noise_variance=observation_noise_variance,
        cholesky_fn=cholesky_fn,
        always_yield_multivariate_normal=True,
    )



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Experimenter factories."""

from typing import Optional, Protocol

import attr
import numpy as np
from vizier._src.benchmarks.experimenters import experimenter
from vizier._src.benchmarks.experimenters import noisy_experimenter
from vizier._src.benchmarks.experimenters import numpy_experimenter
from vizier._src.benchmarks.experimenters import shifting_experimenter
from vizier._src.benchmarks.experimenters.synthetic import bbob


class ExperimenterFactory(Protocol):
  """Abstraction for creating Experimenters."""

  def __call__(self) -> experimenter.Experimenter:
    """Creates the Experimenter."""


@attr.define
class BBOBExperimenterFactory(ExperimenterFactory):
  """Factory for a BBOB function."""

  # Should be a BBOB function name in bbob.py (name should match exactly).
  name: str = attr.field(validator=attr.validators.instance_of(str))
  dim: int = attr.field(
      validator=[attr.validators.instance_of(int),
                 attr.validators.gt(0)])

  def __call__(self) -> numpy_experimenter.NumpyExperimenter:
    bbob_function = getattr(bbob, self.name, None)
    if bbob_function is None:
      raise ValueError(f'{self.name} is not a valid BBOB function in bbob.py')
    return numpy_experimenter.NumpyExperimenter(
        bbob_function, bbob.DefaultBBOBProblemStatement(self.dim))


@attr.define
class SingleObjectiveExperimenterFactory(ExperimenterFactory):
  """Factory for a single objective Experimenter."""

  base_factory: ExperimenterFactory = attr.field()
  # An array of doubles that is broadcastable to dim of search space.
  shift: Optional[np.ndarray] = attr.field(default=None)
  # Should be one of the noise types in noisy_experimenter.py
  noise_type: Optional[str] = attr.field(default=None)

  def __call__(self) -> experimenter.Experimenter:
    """Creates the SingleObjective Experimenter."""
    exptr = self.base_factory()
    if self.shift is not None:
      exptr = shifting_experimenter.ShiftingExperimenter(
          exptr, shift=self.shift)
    if self.noise_type is not None:
      exptr = noisy_experimenter.NoisyExperimenter(
          exptr, noise_type=self.noise_type.upper())
    return exptr



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Random vectorized optimizer to be used in convergence test."""

from typing import Optional

import numpy as np
from vizier._src.algorithms.optimizers import vectorized_base as vb
from vizier.pyvizier import converters


class RandomVectorizedStrategy(vb.VectorizedStrategy):
  """Random vectorized strategy."""

  def __init__(
      self,
      converter: converters.TrialToArrayConverter,
      suggestion_batch_size: int,
      seed: Optional[int] = None,
      prior_features: Optional[np.ndarray] = None,
      prior_rewards: Optional[np.ndarray] = None,
  ):
    self._converter = converter
    self._suggestion_batch_size = suggestion_batch_size
    self._rng = np.random.default_rng(seed)
    self._n_features = sum(
        [spec.num_dimensions for spec in self._converter.output_specs]
    )

  def suggest(self) -> vb.Array:
    return self._rng.uniform(
        low=0,
        high=1,
        size=(
            self._suggestion_batch_size,
            self._n_features,
        ),
    )

  def suggestion_batch_size(self) -> int:
    return self._suggestion_batch_size

  def update(self, rewards: vb.Array) -> None:
    pass


def _random_strategy_factory(
    converter: converters.TrialToArrayConverter,
    suggestion_batch_size: int,
    seed: Optional[int] = None,
    prior_features: Optional[np.ndarray] = None,
    prior_rewards: Optional[np.ndarray] = None,
) -> vb.VectorizedStrategy:
  """Creates a new vectorized strategy based on the Protocol."""
  return RandomVectorizedStrategy(
      converter=converter,
      suggestion_batch_size=suggestion_batch_size,
      seed=seed,
      prior_features=prior_features,
      prior_rewards=prior_rewards,
  )


def create_random_optimizer(
    max_evaluations: int, suggestion_batch_size: int
) -> vb.VectorizedOptimizer:
  """Creates a random optimizer."""
  return vb.VectorizedOptimizer(
      strategy_factory=_random_strategy_factory,
      max_evaluations=max_evaluations,
      suggestion_batch_size=suggestion_batch_size,
  )


def create_random_optimizer_factory() -> vb.VectorizedOptimizerFactory:
  """Creates a random optimizer factory."""
  return vb.VectorizedOptimizerFactory(
      strategy_factory=_random_strategy_factory
  )



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Utils for vectorized eagle strategy."""

from typing import Optional

import attr
import numpy as np
from vizier.pyvizier import converters


@attr.define(kw_only=True)
class EagleParamHandler:
  """Vectorized eagle strategy utils.

  The class is used to account for the different types of Vizier parameters and
  incorporate their naunces into the vectorized eagle strategy.

  Attributes:
    converter: The converter used to convert the original search space. It's
      used to get metadata on the different parameters and attribute the feature
      indices to them.
    config: The eagle strategy configuration used to set the perturbation
      factors for CATEGORICAL features.
    rng: A numpy random generator to generate parameter-aware features, and
      sample categorical features.
    categorical_perturbation_factor: The perturbation factor for categorical
      features.
    pure_categorical_perturbation_factor: The perturbation factor when all
      features are categorical.
    n_features: The total number of feature indices.
    n_categorical: The number of CATEGORICAL associated indices.
    has_categorical: A flag indicating if at least one feature is categorical.
    all_features_categorical: A flag indicating if all features are categorical.
    _categorical_param_mask: A 2D array (n_categorical, n_features) that for
      each categorical parameters (row) has 1s in its associated feature
      indices. The array is used for sampling categorical values.
    _categorical_mask: A 1D array (n_features,) with 1s in the indices of
      categorical features and otherwise 0s. The array is used for sampling
      categorical values.
    _tiebreak_mask: A 1D array (n_features,) with multiplies of epsilons used to
      tie breaking. The array is used for sampling categorical values.
    _oov_mask: A 1D array (n_features,) with 1s in the non-oov indices. The
      array is used to generate random features with 0 value in the OOV indices.
    _epsilon: A small value used in tie-breaker
  """
  converter: converters.TrialToArrayConverter
  categorical_perturbation_factor: float
  pure_categorical_perturbation_factor: float
  rng: np.random.Generator
  # Public variables created by the class
  n_features: int = attr.field(init=False)
  n_categorical: int = attr.field(init=False)
  has_categorical: bool = attr.field(init=False)
  all_features_categorical: bool = attr.field(init=False)
  # Internal variables
  _categorical_params_mask: np.ndarray = attr.field(init=False)
  _categorical_mask: np.ndarray = attr.field(init=False)
  _tiebreak_array: np.ndarray = attr.field(init=False)
  _oov_mask: Optional[np.ndarray] = attr.field(init=False, default=None)
  _epsilon: float = attr.field(init=False, default=1e-5)

  def __attrs_post_init__(self):
    self._init_utils()

  def _init_utils(self):
    """Initialize the utility class and cache arrays for repeated use."""
    valid_types = [
        converters.NumpyArraySpecType.ONEHOT_EMBEDDING,
        converters.NumpyArraySpecType.CONTINUOUS
    ]
    unsupported_params = sum([
        1 for spec in self.converter.output_specs
        if spec.type not in valid_types
    ])
    if unsupported_params:
      raise ValueError('Only CATEGORICAL/CONTINUOUS parameters are supported!')

    self.n_features = sum(
        [spec.num_dimensions for spec in self.converter.output_specs])
    self.n_categorical = sum([
        1 for spec in self.converter.output_specs
        if spec.type == converters.NumpyArraySpecType.ONEHOT_EMBEDDING
    ])
    self.has_categorical = self.n_categorical > 0
    self.all_features_categorical = self.n_features == self.n_categorical
    if self.has_categorical:
      self._create_categorical_masks()

  def _create_categorical_masks(self):
    """Create the categorical masks."""
    self._categorical_params_mask = np.zeros(
        (self.n_categorical, self.n_features))
    self._oov_mask = np.ones((self.n_features,))
    row = 0
    col = 0
    # Create a flag to indicate if the converter uses OOV padding. If none of
    # the CATEGORICAL params use padding the 'oov_mask' is set to None.
    is_pad_oov = False
    for spec in self.converter.output_specs:
      if spec.type == converters.NumpyArraySpecType.ONEHOT_EMBEDDING:
        n_dim = spec.num_dimensions
        self._categorical_params_mask[row, col:col + n_dim] = 1.0
        if spec.num_oovs:
          self._oov_mask[col + n_dim - 1] = 0.0
          is_pad_oov = True
        row += 1
        col += n_dim
      else:
        col += 1
    if not is_pad_oov:
      self._oov_mask = None

    self._tiebreak_array = np.array(
        [-self._epsilon * (i + 1) for i in range(self.n_features)])
    self._categorical_mask = np.sum(self._categorical_params_mask, axis=0)

  def sample_categorical(self, features: np.ndarray) -> np.ndarray:
    """Sample categorical features.

    The categorical sampling is used before returning suggestion to ensure that
    only actual categorical values are suggested. Non categorical features are
    left unchanged.

    For example: If 'features' has one categorical parameter with 3 values and
    one float parameter the conversion will be take the form of:
    (0.1, 0.3, 0.5, 0.4578) -> (0, 1, 0, 0.4578).

    Implementation details:
    ----------------------
    1. For each categorical parameter, isolate its indices in a separate row.
    2. Normalize new row values to probabilities through dividing by the sum.
    3. Create cummulatitive sum of probabilites to generat a CDF.
    4. Add small incremental values to CDF to tie-break (explained more later).
    5. Randomize uniform values for each row to determine which value to sample.
    6. Find the minimum index that its CDF > uniform. See code around
    'sampled_categorical_params' below for more detail and for why we need to
    add the tie-breaking array values.
    7. Flatten each row sampled values and combine with original features.

    Arguments:
      features: (batch_size, n_features)

    Returns:
      The features with sampled categorical parameters. (batch_size ,n_features)
    """
    if not self.has_categorical:
      return features
    batch_size = features.shape[0]
    # Broadcast features to: (batch_size, n_categorical, n_features)
    expanded_shape = (self.n_categorical,) + features.shape
    expanded_features = np.swapaxes(
        np.broadcast_to(features, expanded_shape), 1, 0)
    # Mask each row (which represents a categorical param) to remove values in
    # indices that aren't associated with the parameter indices.
    param_features = expanded_features * self._categorical_params_mask
    # Create probabilities from non-normalized parameter features values.
    probs = param_features / np.sum(param_features, axis=-1, keepdims=True)
    # Generate random uniform values to use for sampling.
    # TODO: Pre-compute random values in batch to improve performance.
    unifs = self.rng.uniform(0.0, 1.0, size=(batch_size, self.n_categorical))
    # Find the locations of the indices that exceed random values.
    locs = np.cumsum(probs, axis=-1) >= np.expand_dims(unifs, axis=-1)
    # Multiply by 'categorical_mask' to mask off cumsum in non-categorical
    # indices, and add 'tiebreak_mask' to find the first index.
    masked_locs = locs * self._categorical_params_mask + self._tiebreak_array
    # Generate the samples so that each parameter features has a single 1 value.
    sampled_categorical_params = np.float_(
        np.int_(masked_locs / np.max(masked_locs, axis=-1, keepdims=True)))
    # Flatten all the categories features to dimension (batch_size, n_features)
    sampled_features = np.sum(sampled_categorical_params, axis=1)
    # Mask categorical features and add the new sampled categorical values.
    return sampled_features + features * (1 - self._categorical_mask)

  def random_features(self, batch_size: int, n_features: int) -> np.ndarray:
    """Create random features with uniform distribution.

    In case there are CATEGORICAL features with OOV we use the 'oov_mask' which
    is 1D numpy array (n_features,) with 0s in OOV indices and otherwise 1s.
    After multiplying (and broadcasting) the randomly generated features with
    the mask we're guaranteed that no features will be created in OOV indices.
    Therefore when mutating fireflies' features (index by index), the final
    suggested features will also have 0s in the OOV indices as desired.

    Arguments:
      batch_size:
      n_features:

    Returns:
      The random features with out of vocabulary indices zeroed out.
    """
    size = (batch_size, n_features)
    if self._oov_mask is not None:
      # Don't create random values for CATEGORICAL features OOV indices.
      # Broadcasting: (batch_size, n_features) x (n_features,)
      return self.rng.uniform(0.0, 1.0, size=size) * self._oov_mask
    else:
      return self.rng.uniform(0.0, 1.0, size=size)

  @property
  def perturbation_factors(self) -> np.ndarray:
    """Create the perturbations factors.

    Returns:
      Array of perturbation factors (n_features,)
    """
    perturbation_factors = []

    if self.all_features_categorical:
      for spec in self.converter.output_specs:
        perturbation_factors.extend(
            [self.pure_categorical_perturbation_factor] * spec.num_dimensions)
    else:
      for spec in self.converter.output_specs:
        if spec.type == converters.NumpyArraySpecType.ONEHOT_EMBEDDING:
          perturbation_factors.extend([self.categorical_perturbation_factor] *
                                      spec.num_dimensions)

        elif spec.type == converters.NumpyArraySpecType.CONTINUOUS:
          perturbation_factors.append(1.0)
    return np.array(perturbation_factors)
